from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional, Tuple, Union

from pygsuite.common.style import Color, TextStyle


class HorizontalAlign(Enum):
    LEFT = "LEFT"
    CENTER = "CENTER"
    RIGHT = "RIGHT"


class VerticalAlign(Enum):
    TOP = "TOP"
    MIDDLE = "MIDDLE"
    BOTTOM = "BOTTOM"


class WrapStrategy(Enum):
    OVERFLOW_CELL = "OVERFLOW_CELL"
    LEGACY_WRAP = "LEGACY_WRAP"
    CLIP = "CLIP"
    WRAP = "WRAP"


class TextDirection(Enum):
    LEFT_TO_RIGHT = "LEFT_TO_RIGHT"
    RIGHT_TO_LEFT = "RIGHT_TO_LEFT"


class HyperlinkDisplayType(Enum):
    LINKED = "LINKED"
    PLAIN_TEXT = "PLAIN_TEXT"


@dataclass
class CellFormat:

    number_format: Optional[dict] = None
    background_color: Optional[Color] = None
    background_color_style: Optional[dict] = None
    borders: Optional[dict] = None
    padding: Optional[dict] = None
    horizontal_alignment: Optional[HorizontalAlign] = None
    vertical_alignment: Optional[VerticalAlign] = None
    wrap_strategy: Optional[WrapStrategy] = None
    text_direction: Optional[TextDirection] = None
    text_format: Optional[TextStyle] = None
    hyperlink_display_type: Optional[HyperlinkDisplayType] = None
    text_rotation: Optional[dict] = None

    def to_json(self) -> Tuple[list, Dict]:  # noqa: C901

        base: Dict[str, Union[str, int, dict, Color]] = {}
        fields_mask = []

        if self.number_format is not None:
            base["numberFormat"] = self.number_format

        if self.background_color is not None:
            base["backgroundColor"] = self.background_color

        if self.background_color_style is not None:
            base["backgroundColorStyle"] = self.background_color_style

        if self.borders is not None:
            base["borders"] = self.borders

        if self.padding is not None:
            base["padding"] = self.padding

        if self.horizontal_alignment is not None:
            base["horizontalAlignment"] = self.horizontal_alignment.value

        if self.vertical_alignment is not None:
            base["verticalAlignment"] = self.vertical_alignment.value

        if self.wrap_strategy is not None:
            base["wrapStrategy"] = self.wrap_strategy.value

        if self.text_direction is not None:
            base["textDirection"] = self.text_direction.value

        if self.text_format is not None:
            masks, base["textFormat"] = self.text_format.to_sheet_style()
            fields_mask.extend(["textFormat." + field for field in masks])

        if self.hyperlink_display_type is not None:
            base["hyperlinkDisplayType"] = self.hyperlink_display_type.value

        if self.text_rotation is not None:
            base["textRotation"] = self.text_rotation

        return fields_mask, base


@dataclass
class Cell:

    user_entered_value: Optional[dict] = None
    user_entered_format: Optional[CellFormat] = None
    hyperlink: Optional[str] = None
    note: Optional[str] = None
    text_format_runs: Optional[dict] = None
    data_validation: Optional[dict] = None
    pivot_table: Optional[dict] = None
    # TODO: should this object contain read-only objects?
    # effective_value: Optional[dict]  # READ ONLY
    # formatted_value: Optional[str]  # READ ONLY
    # effective_format: Optional[CellFormat]  # READ ONLY

    def to_json(self) -> Tuple[str, Dict]:  # noqa: C901

        base: Dict[str, Union[str, dict, int]] = {}
        fields_mask = []

        if self.user_entered_value is not None:
            base["userEnteredValue"] = self.user_entered_value
            fields_mask.append("userEnteredValue")

        if self.user_entered_format is not None:
            masks, base["userEnteredFormat"] = self.user_entered_format.to_json()
            fields_mask.extend(["userEnteredFormat." + field for field in masks])

        if self.hyperlink is not None:
            base["hyperlink"] = self.hyperlink
            fields_mask.append("hyperlink")

        if self.note is not None:
            base["note"] = self.note
            fields_mask.append("note")

        if self.text_format_runs is not None:
            base["textFormatRuns"] = self.text_format_runs
            fields_mask.append("textFormatRuns")

        if self.data_validation is not None:
            base["dataValidation"] = self.data_validation
            fields_mask.append("dataValidation")

        if self.pivot_table is not None:
            base["pivotTable"] = self.pivot_table
            fields_mask.append("pivotTable")

        return ",".join(fields_mask), base



class BaseParagraphElement(object):
    def __init__(self, element: dict, document):
        self._element = element
        self._document = document

    @property
    def start_index(self) -> int:
        return int(self._element.get("startIndex", 0))

    @property
    def end_index(self) -> int:
        return int(self._element.get("endIndex", 0))



"""{
  "startIndex": integer,
  "endIndex": integer,

  // Union field content can be only one of the following:
  "paragraph": {
    object (Paragraph)
  },
  "sectionBreak": {
    object (SectionBreak)
  },
  "table": {
    object (Table)
  },
  "tableOfContents": {
    object (TableOfContents)
  }
  // End of list of possible types for union field content.
}"""

from dataclasses import dataclass
from typing import Dict


@dataclass
class Text:
    base: Dict

    @property
    def text(self):
        base = ""
        for el in self.base.get("textElements") or []:
            if el.get("textRun"):
                base += el.get("textRun")["content"]
        return base


class BaseElement(object):
    def __init__(self, element, document, last):
        self._element = element
        self._document = document
        self._last = last

    @property
    def id(self):
        return self._element["objectId"]

    # @property
    # def position(self):
    #     return (self._element['transform']['scaleX'], self._element['transform']['scaleY'])

    def delete(self):
        reqs = [{"deleteObject": {"objectId": self.id}}]
        self._document._mutation(reqs)

    def update_text(self, text, bullets=False, **kwargs):
        reqs = [
            {"deleteText": {"objectId": self.id, "textRange": {"type": "ALL"}}},
            {"insertText": {"objectId": self.id, "text": text, "insertionIndex": 0}},
        ]

        if bullets and text:
            reqs.append(
                {"createParagraphBullets": {"objectId": self.id, "textRange": {"type": "ALL"}}}
            )

        return self._document._mutation(reqs)

    @property
    def children(self):
        return [self]

    @property
    def end_index(self):
        return self._element.get("endIndex")

    @property
    def start_index(self):
        return self._element.get("startIndex")



from pygsuite.docs.doc_elements import BaseElement
from pygsuite.common.style import TextStyle, ParagraphStyle

from .paragraph_elements import (
    AutoText,
    ColumnBreak,
    HorizontalRule,
    PageBreak,
    TextRun,
    FootnoteReference,
    Equation,
)


# from .doc_elements.image import Image
class ParagraphElement(object):
    def __init__(self, element, document):

        self._element = element
        self._document = document

    def __new__(cls, element, document):
        if element.get("autoText"):
            return AutoText(element, document)
        elif element.get("textRun"):
            return TextRun(element, document)
        elif element.get("ColumnBreak"):
            return ColumnBreak(element, document)
        elif element.get("HorizontalRule"):
            return HorizontalRule(element, document)
        elif element.get("PageBreak"):
            return PageBreak(element, document)
        elif element.get("FootnoteReference"):
            return FootnoteReference(element, document)
        elif element.get("Equation"):
            return Equation(element, document)


class Paragraph(BaseElement):
    def __init__(self, element, document, last):
        BaseElement.__init__(self, element=element, document=document, last=last)
        self._paragraph = self._element.get("paragraph")

    def __repr__(self):
        return f"<Paragraph: {self.text[0:10]}.../>"

    @property
    def elements(self):
        return [ParagraphElement(elem, self._document) for elem in self._paragraph.get("elements")]

    @property
    def text(self):
        return "".join([element.text for element in self.elements if isinstance(element, TextRun)])

    @text.setter
    def text(self, text: str):
        self.delete()
        # TODO: combine with parent?
        self._document._mutation(
            [
                # {
                #     "deleteContentRange": {
                #         "range": {
                #             "segmentId": None,
                #             "startIndex": self.start_index,
                #             "endIndex": self.end_index - 1 if self._last else self.end_index,
                #         }
                #     }
                # },
                {
                    "insertText": {
                        "text": text,
                        "location": {"segmentId": None, "index": self.start_index},
                    }
                }
            ]
        )

    @property
    def style(self):
        return TextStyle.from_doc_style(self._paragraph.get("paragraphStyle"))

    @style.setter
    def style(self, style: ParagraphStyle):
        fields, style = style.to_doc_style()
        self._document._mutation(
            [
                {
                    "updateParagraphStyle": {
                        "range": {
                            "startIndex": self.start_index,
                            "endIndex": self.end_index - 1 if self._last else self.end_index,
                        },
                        "paragraphStyle": style,
                        "fields": fields,
                    }
                }
            ]
        )

    def delete(self):
        end_index = self.end_index - 1 if self._last else self.end_index
        if self.start_index == end_index:
            return
        self._document._mutation(
            [
                {
                    "deleteContentRange": {
                        "range": {
                            "segmentId": None,
                            "startIndex": self.start_index,
                            "endIndex": self.end_index - 1 if self._last else self.end_index,
                        }
                    }
                }
            ]
        )



from pygsuite.docs.doc_elements.paragraph_elements.base_paragraph_element import (
    BaseParagraphElement,
)
from pygsuite.common.style import TextStyle


class TextRun(BaseParagraphElement):
    def __init__(self, element, presentation):
        BaseParagraphElement.__init__(self, element, presentation)
        self._detail = element.get("textRun")

    def __repr__(self):
        return f"<TextRun: {self.text[0:10]}/>"

    @property
    def text(self):
        return self._detail.get("content")

    @text.setter
    def text(self, text: str):
        self.delete()
        # TODO: combine with parent?
        self._document._mutation(
            [
                # {
                #     "deleteContentRange": {
                #         "range": {
                #             "segmentId": None,
                #             "startIndex": self.start_index,
                #             "endIndex": self.end_index - 1 if self._last else self.end_index,
                #         }
                #     }
                # },
                {
                    "insertText": {
                        "text": text,
                        "location": {"segmentId": None, "index": self.start_index},
                    }
                }
            ]
        )

    @property
    def style(self):
        return TextStyle.from_doc_style(self._detail.get("textStyle"))

    @style.setter
    def style(self, style: TextStyle):
        fields, style = style.to_doc_style()
        self._document._mutation(
            [
                {
                    "updateTextStyle": {
                        "range": {"startIndex": self.start_index, "endIndex": self.end_index},
                        "textStyle": style,
                        "fields": fields,
                    }
                }
            ]
        )

    def delete(self):
        end_index = self.end_index
        if self.start_index == end_index:
            return
        self._document._mutation(
            [
                {
                    "deleteContentRange": {
                        "range": {
                            "segmentId": None,
                            "startIndex": self.start_index,
                            "endIndex": self.end_index,
                        }
                    }
                }
            ]
        )



from dataclasses import dataclass

from pygsuite.common.style import Color
from pygsuite.slides.page_elements.common import Text
from .base_element import BaseElement
from typing import Optional


@dataclass
class BackgroundFill:
    state: str
    solid_fill: Color


@dataclass
class ShapeProperties:
    shape_background_fill: Optional[Color] = None
    outline: Optional[str] = None
    shadow: Optional[str] = None
    link: Optional[str] = None
    content_alignment: Optional[str] = None

    # def __post_init__(self):
    #     if isinstance(self.shape_background_fill):

    @classmethod
    def from_api_repr(cls, info: dict):
        if not info:
            return ShapeProperties()
        return ShapeProperties(
            info.get("shapeBackgroundFill"),  # type: ignore
            info.get("outline"),  # type: ignore
            info.get("shadow"),  # type: ignore
            info.get("link"),  # type: ignore
            info.get("contentAlignment"),  # type: ignore
        )

    def to_api_repr(self):
        base = {}
        masks = []

        if self.shape_background_fill is not None:
            base["shapeBackgroundFill"] = {
                "propertyState": "RENDERED",
                "solidFill": self.shape_background_fill.to_slide_style(),
            }
            masks.append("shapeBackgroundFill")
        else:
            base["shapeBackgroundFill"] = {"propertyState": "NOT_RENDERED"}
            masks.append("shapeBackgroundFill")

        if self.outline is not None:
            base["outline"] = self.outline
            masks.append("outline")
        if self.shadow is not None:
            base["shadow"] = self.shadow
            masks.append("shadow")
        if self.link is not None:
            base["link"] = self.link
            masks.append("link")
        masks = ",".join(masks)
        return masks, base


class Shape(BaseElement):
    @classmethod
    def from_id(cls, id, presentation):
        return cls(element={"shape": {}, "objectId": id}, presentation=presentation)

    def __init__(self, element, presentation):
        BaseElement.__init__(self, element, presentation)
        self._details = self._element.get("shape")

    @property
    def id(self):
        return self._element["objectId"]

    def __repr__(self):
        return f"<Shape type:{self.type}>"

    @property
    def type(self):
        return self._details.get("shapeType")

    @property
    def text(self):
        text = self._details.get("text")
        if text:
            return Text(text).text
        else:
            return None

    @text.setter
    def text(self, text: str):
        self.delete_text()
        self.add_text(text)

    def delete_text(self):
        if self.text:
            reqs = [{"deleteText": {"objectId": self.id, "textRange": {"type": "ALL"}}}]
            self._presentation._mutation(reqs=reqs)

    def add_text(self, text, style=None):
        reqs = [{"insertText": {"objectId": self.id, "insertionIndex": 0, "text": text}}]
        if style:
            fields, style = style.to_slide_style()

            reqs.append(
                {
                    "updateTextStyle": {
                        "objectId": self.id,
                        "textRange": {"type": "ALL"},  # {"startIndex": start, "endIndex": end},
                        "style": style,
                        "fields": ",".join(fields),
                    }
                }
            )

        self._presentation._mutation(reqs=reqs)

    @property
    def properties(self):
        return ShapeProperties.from_api_repr(self._details.get("shapeProperties"))

    @properties.setter
    def properties(self, value: ShapeProperties):
        masks, props = value.to_api_repr()
        reqs = [
            {
                "updateShapeProperties": {
                    "objectId": self.id,
                    "fields": masks,
                    "shapeProperties": props,
                }
            }
        ]
        self._presentation._mutation(reqs=reqs)

    @property
    def background_color(self):
        return self.properties.shape_background_fill

    @background_color.setter
    def background_color(self, color: Color):
        new = self.properties
        new.shape_background_fill = color
        self.properties = new

    @property
    def text_style(self):
        return

    @text_style.setter
    def text_style(self, val):
        pass



from enum import Enum
from typing import Optional, Union, List, Dict

import pandas as pd
from googleapiclient.discovery import Resource
from googleapiclient.errors import HttpError

from pygsuite import Clients
from pygsuite.common.drive_object import DriveObject
from pygsuite.common.parsing import parse_id
from pygsuite.enums import FileTypes
from pygsuite.sheets.sheet_properties import SheetProperties
from pygsuite.sheets.worksheet import Worksheet
from pygsuite.utility.decorators import retry


class ValueInputOption(Enum):
    """ValueInputOption: Determines how data should be interpreted.

    Google docs: https://developers.google.com/sheets/api/reference/rest/v4/ValueInputOption
    """

    RAW = "RAW"
    USER_ENTERED = "USER_ENTERED"


class Dimension(Enum):
    """Dimension: Indicates which dimension an operation should apply to.

    Google docs: https://developers.google.com/sheets/api/reference/rest/v4/Dimension
    """

    ROWS = "ROWS"
    COLUMNS = "COLUMNS"


class ValueRenderOption(Enum):
    """ValueRenderOption: Determines how values should be rendered in the output.

    Google docs: https://developers.google.com/sheets/api/reference/rest/v4/ValueRenderOption
    """

    FORMATTED_VALUE = "FORMATTED_VALUE"
    UNFORMATTED_VALUE = "UNFORMATTED_VALUE"
    FORMULA = "FORMULA"


class DateTimeRenderOption(Enum):
    """DateTimeRenderOption: Determines how dates should be rendered in the output.

    Google docs: https://developers.google.com/sheets/api/reference/rest/v4/DateTimeRenderOption
    """

    SERIAL_NUMBER = "SERIAL_NUMBER"
    FORMATTED_STRING = "FORMATTED_STRING"


def create_new_spreadsheet(title: str, client: Optional[Resource] = None):
    """Function to create a new spreadsheet given a client connection to the GSuite API,
       and a title for the new sheet.

    # TODO: add functionality to specify "filepath"?

    Args:
        service (googleapiclient.discovery.Resource): Connection to the Google API Sheets resource.
        title (str): Name for the new spreadsheet document

    Returns:
        id (str): the id of the spreadsheet
    """

    if not isinstance(title, str):
        raise TypeError("The name of the spreadsheet must be given as a string.")

    from pygsuite import Clients

    service = client or Clients.sheets_client

    request = {"properties": {"title": title}}

    spreadsheet = service.spreadsheets().create(body=request, fields="spreadsheetId").execute()
    id = spreadsheet.get("spreadsheetId")

    return Spreadsheet(client=service, id=id)


class ValueResponse(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_list(self):
        return list(self)

    def to_df(self, header: bool):
        """Method to use with self.get_values_from_range(self, cell_range) to return a pandas.DataFrame of values.

        Returns:
            df (pd.DataFrame): A pandas.DataFrame of the cell data gotten from get_values_from_range()
        """
        if header:
            header = self[0]
            records = self[1:]
            df = pd.DataFrame.from_records(data=records, columns=header)
        else:
            records = self
            df = pd.DataFrame.from_records(data=records)

        return df


class Spreadsheet(DriveObject):
    """Base class for the GSuite Spreadsheets API."""

    file_type = FileTypes.SHEETS

    def __init__(self, id: str, client: Optional[Resource] = None):
        """Method to initialize the class.

        The __init__ method accepts a client connection to the Google API, which it uses to retrieve the properties
        of the spreadsheet and create a _spreadsheet object to execute other requests of the API.

        Args:
            id (str): id of the Google Spreadsheet. Spreadsheet id can be found in the Spreadsheet URL between /d/ and /edit, eg:
                https://docs.google.com/spreadsheets/d/<spreadsheetId>/edit#gid=0
            client (googleapiclient.discovery.Resource): connection to the Google API Sheets resource.
        """

        self.service = client or Clients.sheets_client
        self.id = parse_id(id) if id else None
        DriveObject.__init__(self, id=id, client=client)

        self._spreadsheet = self.service.spreadsheets().get(spreadsheetId=self.id).execute()
        self._properties = self._spreadsheet.get("properties")

        # queues to add to and run in flush()
        self._spreadsheets_update_queue: List[Dict] = []
        self._values_update_queue: List[Dict] = []

    @classmethod
    def create_new(cls, title: str, client=None):
        client = client or Clients.sheets_client
        body = {"properties": {"title": title}}
        new = client.spreadsheets().create(body=body).execute()
        return Spreadsheet(id=new.get("spreadsheetId"), client=client)

    def __getitem__(self, key: Union[str, int]):

        # if the key is an integer, use the int as worksheet index
        if isinstance(key, int):
            return self.worksheets[key]
        # if the key is a string, try to match to a worksheet name
        elif isinstance(key, str):
            try:
                return [worksheet for worksheet in self.worksheets if worksheet.name == key][0]
            except IndexError:
                raise KeyError(
                    "No worksheet with the title '{key}' exists. The worksheets included in your spreadsheet are: {worksheets}".format(
                        key=key, worksheets=[worksheet.name for worksheet in self.worksheets]
                    )
                )
        else:
            raise ValueError(
                "The key must be a string or integer. Use a string to get a worksheet by name, and integer to get a worksheet by index."
            )

    @property
    def worksheets(self):
        """List of Worksheet objects for each worksheet in the Spreadsheet."""
        return [Worksheet(sheet, self) for sheet in self._spreadsheet.get("sheets")]

    def refresh(self):
        """Method to refresh the spreadsheets API connection."""
        self._spreadsheet = self.service.spreadsheets().get(spreadsheetId=self.id).execute()

        self._spreadsheets_update_queue = []
        self._values_update_queue = []

    @retry((HttpError), tries=3, delay=10, backoff=5)
    def flush(
        self,
        reverse: bool = False,
        value_input_option: ValueInputOption = ValueInputOption.USER_ENTERED,
        include_values_in_response: bool = False,
        response_value_render_option: ValueRenderOption = ValueRenderOption.FORMATTED_VALUE,
        response_date_time_render_option: DateTimeRenderOption = DateTimeRenderOption.SERIAL_NUMBER,
    ):
        """Method to execute the change queues created through other methods.

        Args:
            reverse (bool): If True, the changes queued will be executed in reverse order.
            value_input_option (ValueInputOption): Indicates which dimension an operation should apply to.
            include_values_in_response (bool): If True, the API response will include the values changed.
            response_value_render_option (ValueRenderOption): Determines how values should be rendered in the output.
            response_date_time_render_option (DateTimeRenderOption): Determines how dates should be rendered in the output.

        Returns:
            response_dict (dict): Dictionary of API response JSON.
        """

        if reverse:
            _spreadsheets_update_queue = list(reversed(self._spreadsheets_update_queue))
            _values_update_queue = list(reversed(self._values_update_queue))
        else:
            _spreadsheets_update_queue = self._spreadsheets_update_queue
            _values_update_queue = self._values_update_queue

        response_dict = dict()

        if len(_spreadsheets_update_queue) > 0:
            response_dict["spreadsheets_update_response"] = (
                self.service.spreadsheets()
                .batchUpdate(body={"requests": _spreadsheets_update_queue}, spreadsheetId=self.id)
                .execute()
                .get("responses")
            )

        if len(_values_update_queue) > 0:
            response_dict["values_update_response"] = (
                self.service.spreadsheets()
                .values()
                .batchUpdate(
                    spreadsheetId=self.id,
                    body={
                        "valueInputOption": value_input_option.value,
                        "data": _values_update_queue,
                        "includeValuesInResponse": include_values_in_response,
                        "responseValueRenderOption": response_value_render_option.value,
                        "responseDateTimeRenderOption": response_date_time_render_option.value,
                    },
                )
                .execute()
                .get("responses")
            )

        # the queues are reset in the refresh() method
        # self._spreadsheets_update_queue = []
        # self._values_update_queue = []
        self.refresh()

        return response_dict

    def create_sheet(self, sheet_properties: Optional[SheetProperties] = None, **kwargs):
        """Add a new sheet to the spreadsheet."""
        sheet_properties = sheet_properties or SheetProperties(**kwargs)
        base = {"addSheet": {"properties": sheet_properties.to_json()}}
        self._spreadsheets_update_queue.append(base)
        return self

    def create_and_return_sheet(
        self, sheet_properties: Optional[SheetProperties] = None, **kwargs
    ) -> "Worksheet":
        before = [sheet.name for sheet in self.worksheets]
        self.create_sheet(sheet_properties=sheet_properties, **kwargs)
        self.flush()
        try:
            new = [sheet.name for sheet in self.worksheets if sheet.name not in before][0]
        except IndexError:
            raise ValueError("Error creating new sheet!")
        return self[new]

    def update_sheet(self, worksheet, sheet_properties):

        pass

    def clear_range(self, range: str):
        self.service.spreadsheets().values().clear(spreadsheetId=self.id, range=range).execute()
        return self

    def delete_sheet(self, title: Optional[str] = None, id: Optional[int] = None):
        """Method to delete a sheet by the title or id.

        Args:
            title (Optional: str): The title of a sheet to delete.
            id (Optional: int): The id of a sheet to delete.

        Returns:
            self (Spreadsheet): The object itself, to method chain.
        """

        # if an id is not given, title must be given; get id from title
        if not id:
            assert title is not None
            id = self.__getitem__(key=title).id

        base = {"deleteSheet": {"sheetId": id}}
        self._spreadsheets_update_queue.append(base)

        return self

    def get_values_from_range(self, cell_range: str) -> ValueResponse:
        """Method to get data from a Spreadsheet range. Returns a value object which can be chained
        to a dataframe or list.

        Args:
            cell_range (str): The range of cells for which to get data.

        Returns:
            self (Spreadsheet): The object itself, to method chain.
        """

        get_response = (
            self.service.spreadsheets()
            .values()
            .batchGet(spreadsheetId=self.id, ranges=[cell_range])
            .execute()
        )

        value_range = get_response.get("valueRanges")[0]  # TODO: cleanup
        values = value_range.get("values", [])
        return ValueResponse(values)

    def to_list(self, cell_range: str) -> list:
        """Method to use to return a list of values.

        Returns:
            self._values (list): A list of the cell data gotten from get_values_from_range()
        """
        return self.get_values_from_range(cell_range=cell_range)

    def to_df(self, cell_range: str, header: bool = True) -> pd.DataFrame:
        """Method to use to return a pandas.DataFrame of values.

        Returns:
            df (pd.DataFrame): A pandas.DataFrame of the cell data gotten from get_values_from_range()
        """
        return self.get_values_from_range(cell_range=cell_range).to_df(header=header)

    def insert_data(
        self, insert_range: str, values: list, major_dimension: Dimension = Dimension.ROWS
    ):
        """Method to insert data from a list into a given range in the Spreadsheet.

        Args:
            insert_range (str): The cell range in which to insert data.
            values (list): The data values, as a list, to insert.
            major_dimension (Dimension): Indicates which dimension an operation should apply to.

        Returns:
            self (Spreadsheet): The object itself, to method chain.
        """

        value_range = {
            "range": insert_range,
            "majorDimension": major_dimension.value,
            "values": values,
        }

        self._values_update_queue.append(value_range)

        return self

    def insert_data_from_df(
        self, df: pd.DataFrame, insert_range: str, major_dimension: Dimension = Dimension.ROWS
    ):
        """Method to insert data from a pd.DataFrame into a given range in the Spreadsheet.

        Args:
            df (pd.DataFrame): Dataframe of data to insert.
            insert_range (str): The cell range in which to insert data.
            major_dimension (Dimension): Indicates which dimension an operation should apply to.

        Returns:
            self (Spreadsheet): The object itself, to method chain.
        """

        # TODO: this insert_data_from_df method might be better as a method that we can
        # attach to insert_data--something like: mySpreadsheet.insert_data().from_df()

        header = df.columns.values.tolist()
        data = df.values.tolist()

        values = []
        if len(header) > 0:
            values.append(header)
        values.extend(data)

        self.insert_data(insert_range=insert_range, values=values, major_dimension=major_dimension)

        return self

    @property
    def url(self):
        return f"https://docs.google.com/spreadsheets/d/{self.id}"



from typing import Optional

from pygsuite import Clients
from pygsuite.enums import FileTypes, PermissionType, UserType

DRIVE_V3_API_URL = "https://www.googleapis.com/drive/v3/files"


def default_callback(request_id, response, exception):
    if exception:
        # Handle error
        raise ValueError(exception)


class Drive:
    def __init__(self, client=None):
        client = client or Clients.drive_client_v3
        self.service = client

    def find_files(self, type: FileTypes, name: Optional[str] = None):
        q = f'mimeType="{type}"'
        if name:
            q += f' and name = "{name}"'
        base = self.service.files().list(q=q).execute()
        files = base.get("files")
        page_token = base.get("nextPageToken")
        while page_token is not None:
            base = self.service.files().list(q=q, page_token=page_token).execute()
            files += base.get("files")
            page_token = base.get("nextPageToken")
        return files

    def update_file_permissions(
        self, file_id, address, role=PermissionType.READER, type=UserType.USER
    ):
        """Deprecate this in favor of the object specific methods"""
        batch = self.service.new_batch_http_request(callback=default_callback)
        user_permission = {"type": type.value, "role": role.value, "emailAddress": address}
        batch.add(
            self.service.permissions().create(fileId=file_id, body=user_permission, fields="id")
        )
        batch.execute()

    def create_file(self, file_id, address, role=PermissionType.READER, type=UserType.USER):
        batch = self.service.new_batch_http_request(callback=default_callback)
        user_permission = {"type": type.value, "role": role.value, "emailAddress": address}
        batch.add(
            self.service.permissions().create(fileId=file_id, body=user_permission, fields="id")
        )
        batch.execute()

    def copy_file(self, file_id, title: str, folder_id: str):
        body = {"name": title, "parents": [folder_id]}
        response = self.service.files().copy(fileId=file_id, body=body).execute()
        return response



from .doc_elements.base_element import BaseElement
from .doc_elements.paragraph import Paragraph
from .doc_elements.section_break import SectionBreak
from .doc_elements.table import Table


# from .doc_elements.image import Image


class DocElement(object):
    def __init__(self, element, document, last: bool = False):
        self._element = element
        self._document = document

    def __new__(cls, element, document, last):
        if element.get("paragraph"):
            return Paragraph(element, document, last)
        elif element.get("table"):
            return Table(element, document, last)
        elif element.get("sectionBreak"):
            return SectionBreak(element, document, last)
        else:
            return BaseElement(element, document, last)

    @property
    def start_index(self) -> int:
        return self._element.get("startIndex")

    @property
    def end_index(self) -> int:
        return self._element.get("endIndex")



from enum import Enum


class PermissionType(Enum):
    OWNER = "owner"
    ORGANIZER = "organizer"
    FILE_ORGANIZER = "fileOrganizer"
    WRITER = "writer"
    COMMENTER = "commenter"
    READER = "reader"


class UserType(Enum):
    USER = "user"
    GROUP = "group"
    DOMAIN = "domain"
    ANYONE = "anyone"


class FileTypes(Enum):
    SHEETS = "application/vnd.google-apps.spreadsheet"
    DOCS = "application/vnd.google-apps.document"
    SLIDES = "application/vnd.google-apps.presentation"
    FOLDER = "application/vnd.google-apps.folder"
    PDF = "application/pdf"
    UNKNOWN = ""

    def __str__(self):
        return self.value



# from math import floor
from string import ascii_letters, ascii_uppercase

from pygsuite.constants import SHEETS_MAX_COLUMN_NUMBER


def index_to_alphabet(input_index: int) -> str:
    """Method to convert integer to A1 notation.

    :type input_index: int
    :param idx: Cell index to convert.

    :rtype: str
    :returns: A1 notation of the cell.
    """

    try:
        assert input_index <= SHEETS_MAX_COLUMN_NUMBER
    except AssertionError:
        raise ValueError(
            f"Input of column index {input_index} exceeds the maximum column number allowed of {SHEETS_MAX_COLUMN_NUMBER}."
        )

    a1_notation = ""

    while input_index > 0:
        input_index, remainder = divmod(input_index - 1, 26)
        a1_notation = ascii_uppercase[remainder] + a1_notation

    return a1_notation


def alphabet_to_index(cell_ref: str) -> int:
    """Method to convert A1 notation to an integer.
    NOTE: This method will return zero-centered indexes,
            despite `index_to_alphabet` taking one-centered index inputs.

    :type cell_ref: str
    :param cell_ref: A1 notation of cell to convert.

    :rtype: int
    :returns: Index notation of the cell.
    """

    idx = 0
    for char in cell_ref:
        if char in ascii_letters:
            idx = idx * 26 + (ord(char.upper()) - ord("A")) + 1

    return idx - 1



from .base_element import BaseElement, Text


class TableCell(BaseElement):
    def __init__(self, element, presentation, table_id, row_index, column_index):
        BaseElement.__init__(self, element, presentation)
        self.table_id = table_id
        self.row_index = row_index
        self.column_index = column_index

    def __repr__(self):
        return f"cell {self.row_index} by {self.column_index} in table {self.table_id}"

    @property
    def text(self):
        if self._element.get("text"):
            return Text(self._element.get("text")).text
        return None

    @property
    def cell_location(self):
        return {"rowIndex": self.row_index, "columnIndex": self.column_index}

    def update_text(self, text, bullets=False, font_size=None, **kwargs):
        # only delete text if there is some
        if self.text:
            reqs = [
                {
                    "deleteText": {
                        "objectId": self.table_id,
                        "cellLocation": self.cell_location,
                        "textRange": {"type": "ALL"},
                    }
                }
            ]
        else:
            reqs = []
        reqs += [
            {
                "insertText": {
                    "objectId": self.table_id,
                    "cellLocation": self.cell_location,
                    "text": text,
                    "insertionIndex": 0,
                }
            }
        ]

        if bullets and text:
            reqs.append(
                {
                    "createParagraphBullets": {
                        "objectId": self.table_id,
                        "cellLocation": self.cell_location,
                        "textRange": {"type": "ALL"},
                    }
                }
            )

        if font_size and text:
            reqs.append(
                {
                    "updateTextStyle": {
                        "objectId": self.table_id,
                        "cellLocation": self.cell_location,
                        "style": {"fontSize": {"magnitude": font_size, "unit": "PT"}},
                        "textRange": {"type": "ALL"},
                        "fields": "fontSize",
                    }
                }
            )

        return self._presentation._mutation(reqs)

    def set_background_color(self, theme_color=None, rgb_color=None):
        assert theme_color or rgb_color
        assert not (theme_color and rgb_color)

        if theme_color:
            color = {"themeColor": theme_color}
        elif rgb_color:
            color = {
                "rgbColor": {
                    "red": rgb_color[0] / 256,
                    "green": rgb_color[1] / 256,
                    "blue": rgb_color[2] / 256,
                }
            }
        else:
            raise ValueError("No color provided")

        reqs = [
            {
                "updateTableCellProperties": {
                    "objectId": self.table_id,
                    "tableRange": {"location": self.cell_location, "rowSpan": 1, "columnSpan": 1},
                    "tableCellProperties": {
                        "tableCellBackgroundFill": {"solidFill": {"color": color, "alpha": 1}}
                    },
                    "fields": "tableCellBackgroundFill.solidFill.color",
                }
            }
        ]
        self._presentation._mutation(reqs)

    def set_border_color(self, theme_color=None, rgb_color=None, weight=2):
        assert theme_color or rgb_color
        assert not (theme_color and rgb_color)

        if theme_color:
            color = {"themeColor": theme_color}
        elif rgb_color:
            color = {
                "rgbColor": {
                    "red": rgb_color[0] / 256,
                    "green": rgb_color[1] / 256,
                    "blue": rgb_color[2] / 256,
                }
            }
        else:
            raise ValueError("No color provided")

        reqs = [
            {
                "updateTableBorderProperties": {
                    "objectId": self.table_id,
                    "tableRange": {"location": self.cell_location, "rowSpan": 1, "columnSpan": 1},
                    "tableBorderProperties": {
                        "tableBorderFill": {"solidFill": {"color": color, "alpha": 1}},
                        "weight": {"magnitude": weight, "unit": "PT"},
                    },
                    "fields": "tableBorderFill.solidFill.color,weight.magnitude",
                }
            }
        ]
        self._presentation._mutation(reqs)


class TableRow(BaseElement):
    def __init__(self, element, presentation, table_id, row_index):
        BaseElement.__init__(self, element, presentation)
        self.table_id = table_id
        self.row_index = row_index

    @property
    def cells(self):
        return [
            TableCell(
                val,
                self._presentation,
                table_id=self.table_id,
                column_index=idx,
                row_index=self.row_index,
            )
            for idx, val in enumerate(self._element["tableCells"])
        ]

    def __getitem__(self, item):
        return self.cells[item]


class Table(BaseElement):
    def __init__(self, element, presentation):
        BaseElement.__init__(self, element, presentation)
        self._details = self._element.get("table")

    def __getitem__(self, item):
        return self.rows[item]

    @property
    def size(self):
        return self.row_count * self.column_count

    @property
    def row_count(self):
        return self._details["rows"]

    @property
    def column_count(self):
        return self._details["columns"]

    @property
    def rows(self):
        return [
            TableRow(row, self._presentation, self.id, idx)
            for idx, row in enumerate(self._details["tableRows"])
        ]

    @property
    def id(self):
        return self._element["objectId"]

    @property
    def children(self):
        out = []
        for row in self.rows:
            for cell in row.cells:
                out.append(cell)
        return out



from typing import List, Optional

from pygsuite.common.style import TextStyle
from pygsuite.docs.doc_elements.paragraph import Paragraph
from pygsuite.docs.element import DocElement


class Body(object):
    def __init__(self, body, document):
        from pygsuite import Document

        self._document: Document = document
        self._pending = []

    @property
    def _body(self):
        return self._document._document.get("body")

    @property
    def content(self) -> List[DocElement]:
        content_len = len(self._body.get("content"))
        return [
            DocElement(element, self._document, idx == content_len - 1)
            for idx, element in enumerate(self._body.get("content"))
        ]

    @property
    def paragraphs(self) -> List[Paragraph]:
        return [item for item in self.content if isinstance(item, Paragraph)]

    def delete(self, flush=True):
        # save the last character of the last element
        for object in reversed(self.content):
            object.delete()
        if flush:
            self._document.flush()
        self.content = []

    @content.setter
    def content(self, x):
        self._body["content"] = x

    @property
    def start_index(self):
        return self.content[0].start_index

    @property
    def end_index(self):
        return self.content[-1].end_index + sum(self._pending)

    @property
    def _end_index_append(self):
        if not self._pending and not self.content:
            return 1
        elif not self.content:
            return sum(self._pending)
        else:
            return self.content[-1].end_index + sum(self._pending)

    def __getitem__(self, item):
        return self.content[item]

    def __setitem__(self, index, value, style=None):
        self.content[index] = value

    def newline(self, count: int = 1):
        self.add_text("\n" * count)

    @property
    def text(self):

        return "".join([element.text for element in self.content if isinstance(element, Paragraph)])

    def flush(self):
        self._document.flush()

    def add_text(self, text: str, position: int = None, style: TextStyle = None):
        # TOOD: add warning
        position = 1 if position == 0 else position
        self._pending.append(len(text))
        if style and not position and self._document._change_queue:
            # if there are pending changes
            # we currently need to flush them to infer proper style positioning
            # TODO: calculate this virtually
            # We could get the size of each pending object, track it, and infer the appropriate style index
            # without flushing
            self._document.flush()
        message = {"insertText": {"text": text}}
        if position is not None:
            message["insertText"]["location"] = {"index": position}
        else:
            message["insertText"]["endOfSegmentLocation"] = {}
        queued = [message]

        if style:
            start = position
            if start is None:
                if not self.content:
                    start = 1
                else:
                    start = self.content[-1].end_index - 1

            end = start + len(text)
            fields, style = style.to_doc_style()
            queued.append(
                {
                    "updateTextStyle": {
                        "range": {"startIndex": start, "endIndex": end},
                        "textStyle": style,
                        "fields": fields,
                    }
                }
            )

        self._document._mutation(queued)

    def add_image(
        self, uri, position: Optional[int] = None, height: Optional[int] = None, width=Optional[int]
    ):
        """Add an image to the document at the specified location."""
        message = {
            "insertInlineImage": {
                "uri": uri,
                # "objectSize": {
                #     object (Size)
                # }
            }
        }
        if height or width:
            size = {}
            if height:
                size["height"] = height
            if width:
                size["width"] = width
            message["insertInlineImage"]["objectSize"] = size
        if position:
            message["insertInlineImage"]["location"] = {"index": position}
        else:
            message["insertInlineImage"]["endOfSegmentLocation"] = {}
        self._document._mutation([message])

    def style(self, style: TextStyle, start: int, end: int):
        # TODO: finish this method
        fields, style = style.to_doc_style()
        message = {
            "updateTextStyle": {
                "textStyle": style,
                "range": {"startIndex": start, "endIndex": end},
                "fields": fields,
            }
        }
        return message



class BaseTextElement(object):
    def __init__(self, element, presentation):
        self._element = element
        self._presentation = presentation

    def end_index(self):
        return self._element.get("endIndex")

    def start_index(self):
        return self._element.get("startIndex")

    def text_run(self):
        return self._element.get("textRun")


class TextRun(BaseTextElement):
    def __init__(self, element, presentation):
        BaseTextElement.__init__(self, element, presentation)
        self._detail = element.get("textRun")

    @property
    def content(self):
        return self._detail.get("content")

    def style(self):
        return self._detail.get("style")


class ParagraphMarker(BaseTextElement):
    def __init__(self, element, presentation):
        BaseTextElement.__init__(self, element, presentation)
        self._detail = element.get("paragraphMarker")

    def style(self):
        return self._detail.get("style")

    def bullet(self):
        return self._detail.get("bullet")


""""textElements": [ {
    "endIndex": 224,
    "paragraphMarker": { "style": {} }
  }, {
    "endIndex": 130,
    "textRun": { "content": "Li lingues differe in li grammatica e li vocabules. Omnicos directe al desirabilite de un nov ", "style": {} }
  }, {
    "endIndex": 143,
    "startIndex": 130,
    "textRun": { "content": "lingua franca", "style": { "italic": True } }
  }, {
    "endIndex": 224,
    "startIndex": 143,
    "textRun": { "content": ": solmen va esser necessi far:\n", "style": {} }
  }, {
    "endIndex": 243,
    "startIndex": 224,
    "paragraphMarker": {
      "style": { "indentStart": { "magnitude": 36, "unit": "PT" }, "direction": "LEFT_TO_RIGHT", "indentFirstLine": { "magnitude": 18, "unit": "PT" }, "spacingMode": "COLLAPSE_LISTS" },
      "bullet": { "listId": "foo123", "glyph": "\u25cf" }
    }
  }, {
    "endIndex": 243,
    "startIndex": 224,
    "textRun": { "content": "uniform grammatica\n", "style": {} }
  }, {
    "endIndex": 257,
    "startIndex": 243,
    "paragraphMarker": {
        "style": { "indentStart": { "magnitude": 36, "unit": "PT" }, "direction": "LEFT_TO_RIGHT", "indentFirstLine": { "magnitude": 18, "unit": "PT" }, "spacingMode": "COLLAPSE_LISTS" },
        "bullet": { "listId": "foo123", "glyph": "\u25cf" }
    }
}, {
    "endIndex": 257,
    "startIndex": 243,
    "textRun": { "content": "Pronunciation\n", "style": {} }
}, {
    "endIndex": 277,
    "startIndex": 257,
    "paragraphMarker": {
        "style": { "indentStart": { "magnitude": 36, "unit": "PT" }, "indentFirstLine": { "magnitude": 18, "unit": "PT" }, "spacingMode": "COLLAPSE_LISTS" },
        "bullet": { "listId": "foo123", "glyph": "\u25cf" }
    }
}, {
    "endIndex": 277,
    "startIndex": 257,
    "textRun": { "content": "plu sommun paroles.\n", "style": {} }
}, {
    "endIndex": 500,
    "startIndex": 277,
    "paragraphMarker": { "style": {} }
}, {
    "endIndex": 500,
    "startIndex": 277,
    "textRun": { "content": "Ka swu thefognay, tay waddeant varpa u inzo.\n", "style": {} }
}]"""



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from pygsuite.slides.page_elements.common import Text
from .base_element import BaseElement


class LineType(Enum):
    STRAIGHT: str
    BENT: str
    CURVED: str
    LINE_CATEGORY_UNSPECIFIED: str


class ArrowStyle(Enum):
    NONE = 0
    STEALTH_ARROW = 1
    FILL_ARROW = 2
    FILL_CIRCLE = 3
    FILL_SQUARE = 4
    FILL_DIAMOND = 5
    OPEN_ARROW = 6
    OPEN_CIRCLE = 7
    OPEN_SQUARE = 8
    OPEN_DIAMOND = 9

    def to_api_repr(self):
        return self.name


@dataclass
class LineConnection:
    object_id: str
    conn_index: int

    def to_api_repr(self):
        return {"connectedObjectId": self.object_id, "connectionSiteIndex": self.conn_index}


@dataclass
class LineProperties:
    line_fill: Optional[str] = None
    weight: Optional[int] = None
    dash_style: Optional[str] = None
    start_arrow: Optional[ArrowStyle] = None
    end_arrow: Optional[ArrowStyle] = None
    link: Optional[str] = None
    start_connection: Optional[LineConnection] = None
    end_connection: Optional[LineConnection] = None

    @classmethod
    def from_api(cls, info):
        return LineProperties(
            info.get("lineFill"),
            info.get("weight"),
            info.get("dashStyle"),
            info.get("startArrow"),
            info.get("endArrow"),
            info.get("link"),
            info.get("startConnection"),
            info.get("endConnection"),
        )

    def to_api_repr(self, line_id):
        fields = []
        properties = {}
        base = {
            "updateLineProperties": {
                "objectId": line_id,
                # "fields": "startConnection",
                # "lineProperties": {"startConnection": conn.to_api_repr()},
            }
        }
        if self.start_connection:
            key = "startConnection"
            fields.append(key)
            properties[key] = self.start_connection.to_api_repr()
        if self.end_connection:
            key = "endConnection"
            fields.append(key)
            properties[key] = self.end_connection.to_api_repr()
        if self.start_arrow:
            key = "startArrow"
            fields.append(key)
            properties[key] = self.start_arrow.to_api_repr()
        if self.start_arrow:
            key = "endArrow"
            fields.append(key)
            properties[key] = self.end_arrow.to_api_repr()
        base["updateLineProperties"]["fields"] = ",".join(fields)
        base["updateLineProperties"]["lineProperties"] = properties
        return base


class Line(BaseElement):
    @classmethod
    def from_id(cls, id, presentation):
        return cls(element={"line": {}, "objectId": id}, presentation=presentation)

    def __init__(self, element, presentation):
        BaseElement.__init__(self, element, presentation)
        self._details = self._element.get("line")

    @property
    def id(self):
        return self._element["objectId"]

    def __repr__(self):
        return f"<Line type:{self.type}>"

    @property
    def type(self):
        return self._details.get("lineType")

    @property
    def category(self):
        return self._details.get("lineCategory")

    @property
    def properties(self):
        return LineProperties(self._details.get("lineProperties"))

    @property
    def start_connection(self):
        base = self._details.get("lineProperties").get("startConnection")
        if not base:
            return None
        else:
            return LineConnection(base.get("connectedObjectId"), base.get("connectionSiteIndex"))

    @start_connection.setter
    def start_connection(self, conn: LineConnection):
        reqs = [
            {
                "updateLineProperties": {
                    "objectId": self.id,
                    "fields": "startConnection",
                    "lineProperties": {"startConnection": conn.to_api_repr()},
                }
            }
        ]
        self._presentation._mutation(reqs=reqs)

    @property
    def end_connection(self):
        base = self._details.get("lineProperties").get("endConnection")
        if not base:
            return None
        else:
            return LineConnection(base.get("connectedObjectId"), base.get("connectionSiteIndex"))

    @end_connection.setter
    def end_connection(self, conn: LineConnection):
        reqs = [
            {
                "updateLineProperties": {
                    "objectId": self.id,
                    "fields": "endConnection",
                    "lineProperties": {"endConnection": conn.to_api_repr()},
                }
            }
        ]
        self._presentation._mutation(reqs=reqs)

    @property
    def text(self):
        text = self._details.get("text")
        if text:
            return Text(text).text
        else:
            return None



from __future__ import print_function

from os.path import dirname, join
import os.path
import pickle

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

from pygsuite.utility.decorators import lazy_property

SCOPES = [
    "https://www.googleapis.com/auth/documents",
    "https://www.googleapis.com/auth/spreadsheets",
    "https://www.googleapis.com/auth/presentations",
    "https://www.googleapis.com/auth/drive",
]

SHEETS_VERSION = "v4"
DOCS_VERSION = "v1"
SLIDES_VERSION = "v1"
DRIVE_VERSION = "v2"


def json_str_to_oauth(token_str: str) -> Credentials:
    import json

    cred_dict = json.loads(token_str)
    return Credentials(**cred_dict)


def get_oauth_credential(credential_string: str) -> Credentials:
    creds = json_str_to_oauth(credential_string)
    if not creds.valid:
        if creds.expired and creds.refresh_token:
            creds.refresh(Request())
            return creds
        raise ValueError(
            "Stored user token is no longer valid and no refresh token! Must be regenerated"
        )
    return creds


class _Clients(object):
    def __init__(self):
        self.cred_path = None
        self.cred_text = None
        self.auth = None

    def validate(self):
        if not self.auth:
            raise ValueError("Need to provide credential path or credential text or auth object.")

    def auth_default(self, project: str = None):
        import google.auth

        self.auth, project_id = google.auth.default(quota_project_id=project)

    def create_client_file_from_string(self):
        from tempfile import TemporaryDirectory
        import os

        with TemporaryDirectory() as temppath:
            with open(os.path.join(temppath, "writer-key.json"), "w") as keyfile:
                keyfile.write(self.cred_text)
            flow = InstalledAppFlow.from_client_secrets_file(keyfile.name, SCOPES)
        return flow

    def authorize(self, auth, project=None):
        self.auth = auth

    def authorize_string(self, auth_string: str):
        self.auth = get_oauth_credential(auth_string)

    def local_file_auth(self, path: str):
        directory = dirname(path)
        pickle_path = join(directory, "cache.pickle")
        creds = None
        if os.path.exists(pickle_path):
            with open(pickle_path, "rb") as token:
                creds = pickle.load(token)
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(path, SCOPES)
                creds = flow.run_local_server(port=0)
            with open(pickle_path, "wb") as token:
                pickle.dump(creds, token)
        self.auth = creds

    @lazy_property
    def docs_client(self):
        self.validate()
        return build("docs", DOCS_VERSION, credentials=self.auth)

    @lazy_property
    def sheets_client(self):
        self.validate()
        return build("sheets", SHEETS_VERSION, credentials=self.auth)

    @lazy_property
    def slides_client(self):
        self.validate()
        return build("slides", SLIDES_VERSION, credentials=self.auth)

    @lazy_property
    def _local_sheets_client(self):
        return build("sheets", SHEETS_VERSION, credentials=self.auth)

    @lazy_property
    def local_docs_client(self):
        return build("docs", DOCS_VERSION, credentials=self.auth)

    @lazy_property
    def _local_slides_client(self):
        return build("slides", SLIDES_VERSION, credentials=self.auth)

    @lazy_property
    def drive_client(self):
        return self.drive_client_v3

    @lazy_property
    def drive_client_v2(self):
        self.validate()
        return build("drive", "v2", credentials=self.auth)

    @lazy_property
    def drive_client_v3(self):
        self.validate()
        return build("drive", "v3", credentials=self.auth)


Clients = _Clients()



from .auto_text import AutoText
from .column_break import ColumnBreak
from .equation import Equation
from .footnote_reference import FootnoteReference
from .horizontal_rule import HorizontalRule
from .inline_object_element import InlineObjectElement
from .page_break import PageBreak
from .text_run import TextRun

__all__ = [
    "TextRun",
    "AutoText",
    "PageBreak",
    "ColumnBreak",
    "FootnoteReference",
    "HorizontalRule",
    "Equation",
    "InlineObjectElement",
]



from pygsuite.slides.presentation import Presentation
from pygsuite.slides.enums import ShapeType
from pygsuite.slides.element_properties import ElementProperties, AnchorPosition
from pygsuite.slides.page_elements.shape import ShapeProperties
from pygsuite.slides.page_elements.line import LineConnection

__all__ = [
    "Presentation",
    "ShapeType",
    "ElementProperties",
    "ShapeProperties",
    "LineConnection",
    "AnchorPosition",
]



class Reply(object):
    """{
        "kind": "drive#commentReply",
        "replyId": string,
        "createdDate": datetime,
        "modifiedDate": datetime,
        "author": {
            "kind": "drive#user",
            "displayName": string,
            "picture": {
                "url": string
            },
            "isAuthenticatedUser": boolean,
            "permissionId": string,
            "emailAddress": string
        },
        "htmlContent": string,
        "content": string,
        "deleted": boolean,
        "verb": string
    }"""

    def __init__(self, _reply):
        self._reply = _reply


class Comment(object):
    """{
        "kind": "drive#comment",
        "selfLink": string,
        "commentId": string,
        "createdDate": datetime,
        "modifiedDate": datetime,
        "author": {
            "kind": "drive#user",
            "displayName": string,
            "picture": {
                "url": string
            },
            "isAuthenticatedUser": boolean,
            "permissionId": string,
            "emailAddress": string
        },
        "htmlContent": string,
        "content": string,
        "deleted": boolean,
        "status": string,
        "context": {
            "type": string,
            "value": string
        },
        "anchor": string,
        "fileId": string,
        "fileTitle": string,
        "replies": [
            replies Resource
        ]
    }"""

    def __init__(self, _comment):
        self._comment = _comment

    @property
    def text(self):
        return self._comment.get("content")

    @property
    def status(self):
        return self._comment.get("status")

    @property
    def author(self):
        return self._comment["author"].get("displayName")

    @property
    def link(self):
        return self._comment.get("selfLink")

    @property
    def replies(self):
        return [Reply(obj) for obj in self._comment.get("replies", []) if not obj.get("deleted")]



from dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class Placeholder:
    base: Dict
    type: Optional[str] = None
    index: Optional[int] = None
    parent_id: Optional[str] = None

    def __post_init__(self):
        self.type = self.base.get("type")
        self.index = self.base.get("index")
        self.parent_id = self.base.get("parentObjectId")

    def __str__(self):
        if self.index:
            return f"{self.type}_{self.index}"
        else:
            return self.type

    @property
    def text(self):
        base = ""
        for el in self.base.get("textElements") or []:
            if el.get("textRun"):
                base += el.get("textRun")["content"]
        return base



def parse_id(input_id: str) -> str:
    if "/" in input_id:
        portions = input_id.split("/")
        for idx, val in enumerate(portions):
            if val == "d":
                return portions[idx + 1]
        raise ValueError("Unable to parse ID from input")
    else:
        return input_id



from typing import Dict, Tuple
from uuid import uuid4


from pygsuite.slides.page_elements.placeholder import Placeholder
from .page_element import PageElement


class Layout(object):
    def __init__(self, layout, presentation):
        self._layout = layout
        self._presentation = presentation

    def __repr__(self):
        return f'Layout<name:"{self.display_name}">'

    @property
    def elements(self):
        return [
            PageElement(element, self._presentation) for element in self._layout.get("pageElements")
        ]

    @property
    def layout_properties(self):
        return self._layout["layoutProperties"]

    @property
    def display_name(self):
        return self.layout_properties["displayName"]

    @property
    def id(self):
        return self._layout["objectId"]

    def _get_smart_placeholder(self, test_str):

        exact = self.placeholders.get(test_str, None)
        if exact:
            return exact
        matches = []
        for key in self.placeholders:
            if test_str in key:
                matches.append(key)
        if len(matches) > 1:
            raise ValueError(
                f"Placeholder key {test_str} was not specific enough - could match any of {matches}"
            )
        if len(matches) == 0:
            raise ValueError(f"No match found for placeholder key {test_str}")
        return self.placeholders[matches[0]]

    def build_reference_map(self, key, val) -> Tuple[Dict, Dict]:

        if "_" in key:
            details = self.placeholders[key]
        else:
            details = self._get_smart_placeholder(key)
        obj_id = str(uuid4()) + key
        return (
            {"insertText": {"objectId": obj_id, "text": val}},
            {
                "layoutPlaceholder": {
                    "type": details.type,
                    "index": details.index,
                    "parentObjectId": details.parent_id,
                },
                "objectId": obj_id,
            },
        )

    @property
    def placeholders(self) -> Dict[str, Placeholder]:
        base = [obj.placeholder for obj in self.elements if obj.placeholder]

        return {str(obj): obj for obj in base}



from enum import Enum


class ShapeType(Enum):
    TYPE_UNSPECIFIED = 0
    TEXT_BOX = 1
    RECTANGLE = 2
    ROUND_RECTANGLE = 3
    ELLIPSE = 4
    ARC = 5
    BENT_ARROW = 6
    BENT_UP_ARROW = 7
    BEVEL = 8
    BLOCK_ARC = 9
    BRACE_PAIR = 10
    BRACKET_PAIR = 11
    CAN = 12
    CHEVRON = 13
    CHORD = 14
    CLOUD = 15
    CORNER = 16
    CUBE = 17
    CURVED_DOWN_ARROW = 18
    CURVED_LEFT_ARROW = 19
    CURVED_RIGHT_ARROW = 20
    CURVED_UP_ARROW = 21
    DECAGON = 22
    DIAGONAL_STRIPE = 23
    DIAMOND = 24
    DODECAGON = 25
    DONUT = 26
    DOUBLE_WAVE = 27
    DOWN_ARROW = 28
    DOWN_ARROW_CALLOUT = 29
    FOLDED_CORNER = 30
    FRAME = 31
    HALF_FRAME = 32
    HEART = 33
    HEPTAGON = 34
    HEXAGON = 35
    HOME_PLATE = 36
    HORIZONTAL_SCROLL = 37
    IRREGULAR_SEAL_1 = 38
    IRREGULAR_SEAL_2 = 39
    LEFT_ARROW = 40
    LEFT_ARROW_CALLOUT = 41
    LEFT_BRACE = 42
    LEFT_BRACKET = 43
    LEFT_RIGHT_ARROW = 44
    LEFT_RIGHT_ARROW_CALLOUT = 45
    LEFT_RIGHT_UP_ARROW = 46
    LEFT_UP_ARROW = 47
    LIGHTNING_BOLT = 48
    MATH_DIVIDE = 49
    MATH_EQUAL = 50
    MATH_MINUS = 51
    MATH_MULTIPLY = 52
    MATH_NOT_EQUAL = 53
    MATH_PLUS = 54
    MOON = 55
    NO_SMOKING = 56
    NOTCHED_RIGHT_ARROW = 57
    OCTAGON = 58
    PARALLELOGRAM = 59
    PENTAGON = 60
    PIE = 61
    PLAQUE = 62
    PLUS = 63
    QUAD_ARROW = 64
    QUAD_ARROW_CALLOUT = 65
    RIBBON = 66
    RIBBON_2 = 67
    RIGHT_ARROW = 68
    RIGHT_ARROW_CALLOUT = 69
    RIGHT_BRACE = 70
    RIGHT_BRACKET = 71
    ROUND_1_RECTANGLE = 72
    ROUND_2_DIAGONAL_RECTANGLE = 73
    ROUND_2_SAME_RECTANGLE = 74
    RIGHT_TRIANGLE = 75
    SMILEY_FACE = 76
    SNIP_1_RECTANGLE = 77
    SNIP_2_DIAGONAL_RECTANGLE = 78
    SNIP_2_SAME_RECTANGLE = 79
    SNIP_ROUND_RECTANGLE = 80
    STAR_10 = 81
    STAR_12 = 82
    STAR_16 = 83
    STAR_24 = 84
    STAR_32 = 85
    STAR_4 = 86
    STAR_5 = 87
    STAR_6 = 88
    STAR_7 = 89
    STAR_8 = 90
    STRIPED_RIGHT_ARROW = 91
    SUN = 92
    TRAPEZOID = 93
    TRIANGLE = 94
    UP_ARROW = 95
    UP_ARROW_CALLOUT = 96
    UP_DOWN_ARROW = 97
    UTURN_ARROW = 98
    VERTICAL_SCROLL = 99
    WAVE = 100
    WEDGE_ELLIPSE_CALLOUT = 101
    WEDGE_RECTANGLE_CALLOUT = 102
    WEDGE_ROUND_RECTANGLE_CALLOUT = 103
    FLOW_CHART_ALTERNATE_PROCESS = 104
    FLOW_CHART_COLLATE = 105
    FLOW_CHART_CONNECTOR = 106
    FLOW_CHART_DECISION = 107
    FLOW_CHART_DELAY = 108
    FLOW_CHART_DISPLAY = 109
    FLOW_CHART_DOCUMENT = 110
    FLOW_CHART_EXTRACT = 111
    FLOW_CHART_INPUT_OUTPUT = 112
    FLOW_CHART_INTERNAL_STORAGE = 113
    FLOW_CHART_MAGNETIC_DISK = 114
    FLOW_CHART_MAGNETIC_DRUM = 115
    FLOW_CHART_MAGNETIC_TAPE = 116
    FLOW_CHART_MANUAL_INPUT = 117
    FLOW_CHART_MANUAL_OPERATION = 118
    FLOW_CHART_MERGE = 119
    FLOW_CHART_MULTIDOCUMENT = 120
    FLOW_CHART_OFFLINE_STORAGE = 121
    FLOW_CHART_OFFPAGE_CONNECTOR = 122
    FLOW_CHART_ONLINE_STORAGE = 123
    FLOW_CHART_OR = 124
    FLOW_CHART_PREDEFINED_PROCESS = 125
    FLOW_CHART_PREPARATION = 126
    FLOW_CHART_PROCESS = 127
    FLOW_CHART_PUNCHED_CARD = 128
    FLOW_CHART_PUNCHED_TAPE = 129
    FLOW_CHART_SORT = 130
    FLOW_CHART_SUMMING_JUNCTION = 131
    FLOW_CHART_TERMINATOR = 132
    ARROW_EAST = 133
    ARROW_NORTH_EAST = 134
    ARROW_NORTH = 135
    SPEECH = 136
    STARBURST = 137
    TEARDROP = 138
    ELLIPSE_RIBBON = 139
    ELLIPSE_RIBBON_2 = 140
    CLOUD_CALLOUT = 141



from dataclasses import dataclass
from enum import Enum
from typing import Optional, Tuple, Dict, Union, Any


@dataclass
class Link:
    url: str
    bookmark_id: str
    heading_id: str


def hex_to_rgb(input: str):
    if input.startswith("#"):
        input = input[1:]
    return tuple(int(input[i : i + 2], 16) / 255.0 for i in (0, 2, 4))


def doc_color_to_color(info: Optional[dict]):
    if not info:
        return None
    base = info.get("color")
    if not base:
        return None
    base = base.get("rgbColor")
    if not base:
        return
    return Color(red=base["red"], green=base["green"], blue=base["blue"])


def doc_link_to_link(info: dict):
    if not info:
        return None
    # TODO: identify link format for these:
    # info.get('bookmarkId'), info.get('headingId')
    return info.get("url")  # Link(info.get('url'), info.get('bookmarkId'), info.get('headingId'))


@dataclass
class Color:
    hex: Optional[str] = None
    red: Optional[float] = None
    blue: Optional[float] = None
    green: Optional[float] = None
    alpha: Optional[float] = None

    @classmethod
    def parse(cls, item: Any):
        if isinstance(item, Color):
            return item
        elif isinstance(item, str):
            return Color(item)
        elif isinstance(item, tuple):
            return Color(None, *item)
        elif not item:
            return None
        else:
            raise ValueError(f"Cannot parse {item} to Color")

    def __post_init__(self):

        assert self.hex or (
            self.red is not None and self.blue is not None and self.green is not None
        )
        if self.hex:
            self.red, self.green, self.blue = hex_to_rgb(self.hex)

    def to_json(self):

        return {"color": {"rgbColor": {"red": self.red, "green": self.green, "blue": self.blue}}}

    def to_sheet_style(self):

        base = {"red": self.red, "green": self.green, "blue": self.blue}

        if self.alpha is not None:
            base["alpha"] = self.alpha

        return base

    def to_slide_style(self):

        base = {"color": {"rgbColor": {"red": self.red, "green": self.green, "blue": self.blue}}}

        if self.alpha is not None:
            base["alpha"] = self.alpha

        return base


@dataclass
class TextStyle:
    font_size: Optional[int] = None
    font: Optional[str] = None
    font_weight: Optional[int] = None
    color: Optional[Color] = None
    background_color: Optional[Color] = None
    bold: Optional[bool] = None
    italic: Optional[bool] = None
    strikethrough: Optional[bool] = None
    small_caps: Optional[bool] = None
    underline: Optional[bool] = None
    link: Optional[str] = None
    font_unit: Optional[str] = "PT"

    def to_doc_style(self) -> Tuple[str, Dict]:  # noqa: C901
        base: Dict[str, Union[bool, str, int, dict]] = {}
        masks = []
        if self.font_size is not None:
            base["fontSize"] = {"magnitude": self.font_size, "unit": "PT"}
            masks.append("fontSize")
        if self.bold is not None:
            base["bold"] = self.bold
            masks.append("bold")
        if self.italic is not None:
            base["italic"] = self.italic
            masks.append("italic")
        if self.underline is not None:
            base["underline"] = self.underline
            masks.append("underline")
        if self.small_caps is not None:
            base["smallCaps"] = self.small_caps
            masks.append("smallCaps")
        if self.background_color is not None:
            base["backgroundColor"] = self.background_color.to_json()
            masks.append("backgroundColor")
        if self.color is not None:
            base["foregroundColor"] = self.color.to_json()
            masks.append("foregroundColor")
        if self.font is not None:
            if not base.get("weightedFontFamily"):
                base["weightedFontFamily"] = {}
            base["weightedFontFamily"]["fontFamily"] = self.font  # type: ignore
            masks.append("weightedFontFamily.fontFamily")
        if self.font_weight is not None:
            if not base.get("weightedFontFamily"):
                base["weightedFontFamily"] = {}
                base["weightedFontFamily"]["fontFamily"] = "Arial"  # type: ignore
            base["weightedFontFamily"]["weight"] = self.font_weight  # type: ignore
            masks.append("weightedFontFamily.fontFamily")
            masks.append("weightedFontFamily.weight")

        return ",".join(masks), base

    def to_sheet_style(self) -> Tuple[list, Dict]:  # noqa: C901

        base: Dict[str, Union[str, bool, int, dict]] = {}
        masks = []

        if self.font_size is not None:
            base["fontSize"] = {"magnitude": self.font_size, "unit": "PT"}
            masks.append("fontSize")
        if self.bold is not None:
            base["bold"] = self.bold
            masks.append("bold")
        if self.italic is not None:
            base["italic"] = self.italic
            masks.append("italic")
        if self.underline is not None:
            base["underline"] = self.underline
            masks.append("underline")
        if self.color is not None:
            base["foregroundColor"] = self.color.to_sheet_style()
            masks.append("foregroundColor")
        if self.font is not None:
            base["fontFamily"] = self.font
            masks.append("fontFamily")

        return masks, base

    def to_slide_style(self):

        base = {}
        masks = []

        if self.font_size is not None:
            base["fontSize"] = {"magnitude": self.font_size, "unit": "PT"}
            masks.append("fontSize")
        if self.bold is not None:
            base["bold"] = self.bold
            masks.append("bold")
        if self.italic is not None:
            base["italic"] = self.italic
            masks.append("italic")
        if self.underline is not None:
            base["underline"] = self.underline
            masks.append("underline")
        if self.color is not None:
            base["foregroundColor"] = self.color.to_sheet_style()
            masks.append("foregroundColor")
        if self.font is not None:
            base["fontFamily"] = self.font
            masks.append("fontFamily")
        if self.link is not None:
            base["link"] = {"url": self.link}
            masks.append("link")
        return masks, base

    @classmethod
    def from_doc_style(cls, info):
        if info.get("namedStyleType"):
            return DefaultFonts[info.get("namedStyleType")]
        return TextStyle(
            font_size=info.get("fontSize", {}).get("magnitude"),
            font_unit=info.get("fontSize", {}).get("unit"),
            font=info.get("font"),
            font_weight=info.get("fonWeight"),
            color=doc_color_to_color(info.get("color")),
            background_color=doc_color_to_color(info.get("backgroundColor")),
            bold=info.get("bold"),
            italic=info.get("italic"),
            strikethrough=info.get("strikethrough"),
            small_caps=info.get("smallCaps"),
            underline=info.get("underline"),
            link=doc_link_to_link(info.get("link")),
        )


class BorderStyle(Enum):
    NONE = "NONE"
    DOTTED = "DOTTED"
    DASHED = "DASHED"
    SOLID = "SOLID"
    SOLID_MEDIUM = "SOLID_MEDIUM"
    SOLID_THICK = "SOLID_THICK"
    DOUBLE = "DOUBLE"


class BorderPosition(Enum):
    TOP = "top"
    BOTTOM = "bottom"
    LEFT = "left"
    RIGHT = "right"
    INNER_HORIZONTAL = "innerHorizontal"
    INNER_VERTICAL = "innerVertical"


@dataclass
class Border:
    """Dataclass to represent a Border object for a border along a cell.

    Google documentation: https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#Border
    """

    BORDER_STYLES = ["NONE", "DOTTED", "DASHED", "SOLID", "SOLID_MEDIUM", "SOLID_THICK", "DOUBLE"]
    COLOR_STYLES = ["rgbColor", "themeColor"]
    THEME_COLOR_TYPES = [
        "TEXT",
        "BACKGROUND",
        "ACCENT1",
        "ACCENT2",
        "ACCENT3",
        "ACCENT4",
        "ACCENT5",
        "ACCENT6",
        "LINK",
    ]

    position: str
    style: BorderStyle
    color: Color
    color_style: Optional[Union[str, Color]] = None

    def to_json(self):

        # assert self.style in self.BORDER_STYLES

        base = {"style": self.style.value, "color": self.color.to_sheet_style()}

        if self.color_style is not None:
            assert self.color_style in self.COLOR_STYLES
            if self.color_style == "rgbColor":
                assert isinstance(self.color_style, Color)
                base["colorStyle"] = self.color_style.to_sheet_style()
            elif self.color_style == "themeColor":
                assert self.color_style in self.THEME_COLOR_TYPES
                base["themeColor"] = self.color_style

        return base


class DefaultFonts(Enum):
    NORMAL_TEXT = TextStyle(font_size=11, font="arial", color=Color(hex="#000000"))
    TITLE = TextStyle(font_size=26, font="arial", color=Color(hex="#000000"))
    SUBTITLE = TextStyle(font_size=15, font="arial", color=Color(hex="#666666"))
    HEADING1 = TextStyle(font_size=20, font="arial", color=Color(hex="#000000"))
    HEADING2 = TextStyle(font_size=16, font="arial", color=Color(hex="#666666"))
    HEADING3 = TextStyle(font_size=14, font="arial", color=Color(hex="#434343"))
    HEADING4 = TextStyle(font_size=12, font="arial", color=Color(hex="#666666"))
    HEADING5 = TextStyle(font_size=11, font="arial", color=Color(hex="#666666"))
    HEADING6 = TextStyle(font_size=11, font="arial", italic=True, color=Color(hex="#666666"))

    def to_doc_style(self):
        return self.value.to_doc_style()


@dataclass
class ParagraphStyle:
    alignment: str
    line_space: int
    direction: str
    spacing_mode: str
    space_above: str
    space_below: str
    border_between: str
    border_top: BorderStyle
    keep_lines_together: bool
    keep_with_next: bool
    avoid_widow_and_orphan: bool
    shading: str

    def to_doc_style(self) -> Tuple[str, Dict]:  # noqa: C901
        base = {}
        masks = []
        if self.alignment is not None:
            base["alignment"] = {"alignment": self.alignment}
            masks.append("alignment")
        # TODO: validate all of these
        for attr in [
            "line_space",
            "direction",
            "spacing_mode",
            "space_above",
            "space_below",
            "border_between",
            "border_top",
            "keep_lines",
            "keep_with_next",
            "avoid_widow_and_orphan",
            "shading",
        ]:
            test = getattr(self, attr)
            if test:
                components = attr.split("_")
                name = components[0] + "".join([component.title() for component in components])
                base[name] = {name: test}
                masks.append(name)
        return ",".join(masks), base

    def to_sheet_style(self) -> Tuple[list, Dict]:  # noqa: C901
        raise NotImplementedError

    @classmethod
    def from_doc_style(cls, info):
        if info.get("namedStyleType"):
            return DefaultFonts[info.get("namedStyleType")]
        return ParagraphStyle(**info)



from dataclasses import dataclass
from typing import Dict

from pygsuite.slides.page_elements.placeholder import Placeholder


@dataclass
class Text:
    base: Dict

    @property
    def text(self):
        base = ""
        for el in self.base.get("textElements") or []:
            if el.get("textRun"):
                base += el.get("textRun")["content"]
        return base


class BaseElement(object):
    def __init__(self, element, presentation):
        self._element = element
        self._presentation = presentation
        self._details = {}

    @property
    def id(self):
        return self._element["objectId"]

    @property
    def position(self):
        return (self._element["transform"]["scaleX"], self._element["transform"]["scaleY"])

    def delete(self):
        reqs = [{"deleteObject": {"objectId": self.id}}]
        self._presentation._mutation(reqs)

    def update_text(self, text, bullets=False, **kwargs):
        reqs = [
            {"deleteText": {"objectId": self.id, "textRange": {"type": "ALL"}}},
            {"insertText": {"objectId": self.id, "text": text, "insertionIndex": 0}},
        ]

        if bullets and text:
            reqs.append(
                {"createParagraphBullets": {"objectId": self.id, "textRange": {"type": "ALL"}}}
            )

        return self._presentation._mutation(reqs)

    @property
    def children(self):
        return [self]

    @property
    def placeholder(self):
        check = self._details.get("placeholder")
        if check:
            return Placeholder(check)



from pygsuite.auth.authorization import Clients
from pygsuite.common.style import DefaultFonts, TextStyle, Color
from pygsuite.docs import Document
from pygsuite.sheets import Spreadsheet
from pygsuite.slides import Presentation
from pygsuite.images import ImageUploader
from pygsuite.drive import Drive
from pygsuite.enums import UserType, PermissionType

__version__ = "0.0.15"
__author__ = "Ethan Dickinson <ethan.dickinson@gmail.com>"
__all__ = [
    "Clients",
    "DefaultFonts",
    "Spreadsheet",
    "Document",
    "Presentation",
    "TextStyle",
    "Color",
    "ImageUploader",
    "Drive",
    "UserType",
    "PermissionType",
]  # type: ignore



""" "elementGroup": {
    object (Group)
  },
  "shape": {
    object (Shape)
  },
  "image": {
    object (Image)
  },
  "video": {
    object (Video)
  },
  "line": {
    object (Line)
  },
  "table": {
    object (Table)
  },
  "wordArt": {
    object (WordArt)
  },
  "sheetsChart": {
    object (SheetsChart)
  }"""



from .base_element import BaseElement
from dataclasses import dataclass

CENTER_INSIDE = "CENTER_INSIDE"
CENTER_CROP = "CENTER_CROP"


@dataclass
class CropProperties:
    left_offset: float
    right_offset: float
    top_offset: float
    bottom_offset: float
    angle: float


@dataclass
class ImageProperties:
    crop_properties: str
    transparency: float
    brightness: float
    contrast: float
    recolor: str
    outline: str
    shadow: str
    link: str


class Image(BaseElement):
    @classmethod
    def from_id(cls, id, presentation):
        return cls(element={"image": {}, "objectId": id}, presentation=presentation)

    def __init__(self, element, presentation):
        BaseElement.__init__(self, element, presentation)
        self._element = element
        self._presentation = presentation
        self._details = self._element.get("image")
        self.text = None

    @property
    def id(self):
        return self._element["objectId"]

    @property
    def position(self):
        return (self._element["transform"]["scaleX"], self._element["transform"]["scaleY"])

    def delete(self):
        reqs = [{"deleteObject": {"objectId": self.id}}]
        self._presentation._mutation(reqs)

    def update_image(self, new_url, replace_method=CENTER_INSIDE):
        reqs = [
            {
                "replaceImage": {
                    "imageObjectId": self.id,
                    "imageReplaceMethod": replace_method,
                    "url": new_url,
                }
            }
        ]

        return self._presentation._mutation(reqs)



from collections import OrderedDict


class KeyList(list):
    def __init__(self, keys, values):
        self.key_dict = OrderedDict()
        for val in zip(keys, values):
            self.key_dict[val[0]] = val[1]
        super(KeyList, self).__init__(values)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.key_dict[item]
        else:
            return super().__getitem__(item)



import time
from functools import wraps
from types import MethodType


class lazy_property(object):
    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.fget(obj)
        setattr(obj, self.func_name, value)

        setattr(obj, "_method_{}".format(self.func_name), MethodType(self.fget, obj))
        return value


def safe_property(f):
    @property
    def wrapper(self, *args, **kwargs):
        if len(self._change_queue) > 0:
            self.flush()
        return f(self, *args, **kwargs)

    return wrapper


def retry(exceptions, tries=4, delay=3, backoff=2, logger=None):
    """
    Retry calling the decorated function using an exponential backoff.

    Args:
        exceptions: The exception to check. may be a tuple of
            exceptions to check.
        tries: Number of times to try (not retry) before giving up.
        delay: Initial delay between retries in seconds.
        backoff: Backoff multiplier (e.g. value of 2 will double the delay
            each retry).
        logger: Logger to use. If None, print.
    """

    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except exceptions as e:
                    msg = "{}, Retrying in {} seconds...".format(e, mdelay)
                    if logger:
                        logger.warning(msg)
                    else:
                        print(msg)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry

    return deco_retry



from dataclasses import dataclass


@dataclass
class Link:
    url: str
    bookmark_id: str
    heading_id: str

    def render(self) -> dict:
        out = {}
        if self.url:
            out = {"url": self.url}
        elif self.bookmark_id:
            out = {"bookmark_id": self.bookmark_id}
        elif self.heading_id:
            out = {"heading_id": self.heading_id}
        return out



from pygsuite.docs.doc_elements import BaseElement


class Footnotes(BaseElement):
    def __init__(self, element, document, last):
        BaseElement.__init__(self, element=element, document=document, last=last)
        self._paragraph = self._element.get("table_of_contents")

    @property
    def section_style(self):
        return self._element.get("sectionStyle")

    def delete(self):
        if not self.start_index:
            return
        if self.start_index == self.end_index:
            return
        self._document._mutation(
            [
                {
                    "deleteContentRange": {
                        "range": {
                            "segmentId": None,
                            "startIndex": self.start_index,
                            "endIndex": self.end_index,
                        }
                    }
                }
            ]
        )



from .page_elements.base_element import BaseElement
from .page_elements.image import Image
from .page_elements.shape import Shape
from .page_elements.table import Table
from .page_elements.line import Line


class PageElement(object):
    def __init__(self, element, presentation):
        self._element = element
        self._presentation = presentation

    def __new__(cls, element, presentation):
        if element.get("shape"):
            return Shape(element, presentation)
        elif element.get("table"):
            return Table(element, presentation)
        elif element.get("image"):
            return Image(element, presentation)
        elif element.get("line"):
            return Line(element, presentation)
        else:
            return BaseElement(element, presentation)



from dataclasses import dataclass
from enum import Enum
from typing import Union, Optional


class AnchorPosition(Enum):
    TOP_LEFT = 1
    CENTER = 2


class MeasurementUnit(Enum):
    PT = 1
    EMU = 2


MARGIN = 0.95
PT_TO_INCH = 72

# DEFAULT SLIDE RATIO
# TODO: genericize
SLIDE_WIDTH = PT_TO_INCH * 10
SLIDE_HEIGHT = PT_TO_INCH * 5.63


def process_dimension(dim, FULL):
    if isinstance(dim, str):
        if dim.endswith("%"):
            return int(dim[0:-1]) * 0.01 * FULL
        else:
            return int(dim)
    else:
        return dim


@dataclass
class ElementProperties:
    x: Union[int, str, float] = 0
    y: Union[int, str, float] = 0
    width: Optional[Union[int, str, float]] = None
    height: Optional[Union[int, str, float]] = None
    object_id: str = ""
    unit_type: str = "PT"
    anchor: AnchorPosition = AnchorPosition.TOP_LEFT

    def __post_init__(self):
        self.x = process_dimension(self.x, SLIDE_WIDTH)
        self.y = process_dimension(self.y, SLIDE_HEIGHT)
        self.width = process_dimension(self.width, SLIDE_WIDTH)
        self.height = process_dimension(self.height, SLIDE_HEIGHT)
        if self.anchor == AnchorPosition.CENTER:
            self.x = self.x - self.width / 2
            self.y = self.y - self.height / 2

    def to_slides_json(self, page_id):
        base = {
            "pageObjectId": page_id,
            "transform": {
                "scaleX": 1,
                "scaleY": 1,
                # "shearX": 1,
                # "shearY": 1,
                "translateX": self.x,
                "translateY": self.y,
                "unit": self.unit_type,
            },
        }

        if self.width or self.height:
            if not self.width and self.height:
                raise ValueError("Must set both height and width when setting either")
            base["size"] = {}
        if self.width:
            base["size"]["width"] = {"magnitude": self.width, "unit": self.unit_type}
        if self.height:
            base["size"]["height"] = {"magnitude": self.height, "unit": self.unit_type}

        return base

    @property
    def base_width(self):
        return SLIDE_WIDTH

    @property
    def base_height(self):
        return SLIDE_HEIGHT



from dataclasses import dataclass
from typing import Dict


@dataclass
class Text:
    base: Dict

    @property
    def text(self):
        base = ""
        for el in self.base.get("textElements") or []:
            if el.get("textRun"):
                base += el.get("textRun")["content"]
        return base



from pygsuite.docs.doc_elements.paragraph_elements.base_paragraph_element import (
    BaseParagraphElement,
)


class FootnoteReference(BaseParagraphElement):
    def __init__(self, element, presentation):
        BaseParagraphElement.__init__(self, element, presentation)
        self._detail = element.get("textRun")

    @property
    def content(self):
        return self._detail.get("content")

    def style(self):
        return self._detail.get("style")



import datetime
import json
from typing import TYPE_CHECKING
from typing import Union, Dict, Tuple
from uuid import uuid4

if TYPE_CHECKING:
    from google.cloud.storage import Client

_SERVICE_ACCOUNT_TYPE = "service_account"


def generate_download_signed_url_v4(bucket, blob_name: str, timeout: int = 15):
    """Generates a v4 signed URL for downloading a blob.

    Note that this method requires a service account key file. You can not use
    this if you are using Application Default Credentials from Google Compute
    Engine or from the Google Cloud SDK.
    """
    # bucket_name = 'your-bucket-name'
    # blob_name = 'your-object-name'

    blob = bucket.blob(blob_name)

    url = blob.generate_signed_url(
        version="v4",
        # This URL is valid for 15 minutes
        expiration=datetime.timedelta(minutes=timeout),
        # Allow GET requests using this URL.
        method="GET",
    )
    return url


class ImageUploader:
    def __init__(self, bucket: str, account_info: Union["Client", str, Dict], timeout: int = 15):
        from google.cloud.storage import Client

        self.account_info = account_info
        self.timeout = timeout
        self._client, self._project = (
            (account_info, None)
            if isinstance(account_info, Client)
            else self._generate_client(account_info)
        )

        self.client = Client(credentials=self._client, project=self._project)
        self.bucket = self.client.bucket(bucket)

    def _generate_client(self, info: Union[Dict, str]) -> Tuple["Client", str]:
        if not isinstance(info, dict):
            info_dict = json.loads(info)
        else:
            info_dict = info
        # The type key should indicate that the file is either a service account
        # credentials file or an authorized user credentials file.
        credential_type = info_dict.get("type")
        if credential_type != _SERVICE_ACCOUNT_TYPE:
            raise ValueError(
                f'Invalid credential type "{credential_type}". Generating signed URLs requires a service account with appropriate permissions.'
            )

        from google.oauth2 import service_account

        return service_account.Credentials.from_service_account_info(info_dict), info_dict.get(
            "project_id"
        )

    def _get_signed_url(self, obj: str):
        return generate_download_signed_url_v4(self.bucket, obj, self.timeout)

    def signed_url_from_file(self, path: str):
        temp_file_name = str(uuid4())
        blob = self.bucket.blob(temp_file_name)
        blob.upload_from_filename(path)
        return self._get_signed_url(temp_file_name)

    def signed_url_from_string(self, string, type: str = "png"):
        temp_file_name = str(uuid4())
        blob = self.bucket.blob(temp_file_name)
        blob.upload_from_string(string, content_type=f"application/{type}")
        return self._get_signed_url(temp_file_name)



from googleapiclient.errors import HttpError

from pygsuite import Clients
from pygsuite.common.drive_object import DriveObject
from pygsuite.common.parsing import parse_id
from pygsuite.docs.body import Body
from pygsuite.docs.footers import Footers
from pygsuite.docs.footnotes import Footnotes
from pygsuite.docs.headers import Headers
from pygsuite.enums import FileTypes
from pygsuite.utility.decorators import retry


class Document(DriveObject):
    """A document on google drive. """

    file_type = FileTypes.DOCS

    def __init__(self, id=None, name=None, client=None, _document=None, local=False):

        if not local:
            client = client or Clients.docs_client
        self.service = client
        DriveObject.__init__(self, id=parse_id(id) if id else None, client=client)
        self._document = _document or client.documents().get(documentId=self.id).execute()
        self._change_queue = []
        self.auto_sync = False

    @classmethod
    def create_new(cls, title: str, client=None):
        client = client or Clients.docs_client
        body = {"title": title}
        new = client.documents().create(body=body).execute()
        return Document(id=new.get("documentId"), client=client)

    def id(self):
        return self._document["id"]

    def _mutation(self, reqs, flush=False):
        if not reqs:
            return None
        self._change_queue += reqs
        if flush or self.auto_sync:
            return self.flush()

    @retry(HttpError, tries=3, delay=5, backoff=3)
    def flush(self, reverse=False):
        if reverse:
            base = reversed(self._change_queue)
        else:
            base = self._change_queue
        final = []
        for item in base:
            if isinstance(item, list):
                for i in item:
                    final.append(i)
            else:
                final.append(item)
        if not base:
            return []
        out = (
            self.service.documents()
            .batchUpdate(body={"requests": final}, documentId=self.id)
            .execute()["replies"]
        )

        self._change_queue = []
        self.refresh()
        return out

    # def delete(self, start=0, end=None, flush=True):
    #     end = end or self.body.end_index
    #     self._mutation([{'deleteContentRange': {'range': {
    #         "segmentId": None,
    #         "startIndex": start,
    #         "endIndex": end
    #     }}}])
    #     if flush:
    #         self.flush()

    @property
    def body(self):
        return Body(self._document.get("body"), self)

    @property
    def footers(self):
        return [Footers(item, self) for item in self._document.get("footers")]

    @property
    def footnotes(self):
        return [Footnotes(item, self) for item in self._document.get("footnotes")]

    @property
    def headers(self):
        return [Headers(item, self) for item in self._document.get("headers")]

    @property
    def title(self):
        return self._document.get("title")

    @title.setter
    def title(self, title: str):
        raise NotImplementedError

    def refresh(self):
        self._document = self.service.documents().get(documentId=self.id).execute()
        self.body._pending = []

    @property
    def url(self):
        return f"https://docs.google.com/docs/d/{self.id}"



# type: ignore
import ast
import re

import setuptools

_version_re = re.compile(r"__version__\s+=\s+(.*)")
with open("pygsuite/__init__.py", "rb") as f:
    _match = _version_re.search(f.read().decode("utf-8"))
    if _match is None:
        print("No version found")
        raise SystemExit(1)
    version = str(ast.literal_eval(_match.group(1)))

with open("requirements.txt", "r") as f:
    install_requires = [line.strip() for line in f.readlines()]

setuptools.setup(
    name="pygsuite",
    version=version,
    url="https://github.com/greenmtnboy/pygsuite",
    author="greenmtnboy",
    author_email="ethan.dickinson@gmail.com",
    description="Package for working with the Gsuite set of tools.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[install_requires],
    extras_require={'images': ['google-cloud-storage', 'pyopenssl']},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from .sheet import Spreadsheet, ValueInputOption, Dimension, ValueRenderOption, DateTimeRenderOption
from .worksheet import Worksheet
from .sheet_properties import SheetType, GridProperties, SheetProperties
from .cell import (
    HorizontalAlign,
    VerticalAlign,
    WrapStrategy,
    TextDirection,
    HyperlinkDisplayType,
    CellFormat,
    Cell,
)


__all__ = [
    "Spreadsheet",
    "ValueInputOption",
    "Dimension",
    "ValueRenderOption",
    "DateTimeRenderOption",
    "Worksheet",
    "SheetType",
    "GridProperties",
    "SheetProperties",
    "HorizontalAlign",
    "VerticalAlign",
    "WrapStrategy",
    "TextDirection",
    "HyperlinkDisplayType",
    "CellFormat",
    "Cell",
]



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from pygsuite.common.style import Color


class SheetType(Enum):
    """SheetType: The kind of sheet.

    Google Docs: https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#SheetType
    """

    GRID = "GRID"
    OBJECT = "OBJECT"


@dataclass
class GridProperties:
    """GridProperties: Properties of a grid.

    Google Docs: https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets#GridProperties
    """

    row_count: Optional[int] = None
    column_count: Optional[int] = None
    frozen_row_count: Optional[int] = None
    frozen_column_count: Optional[int] = None
    hide_gridlines: Optional[bool] = None
    row_group_control_after: Optional[bool] = None
    colum_group_control_after: Optional[bool] = None

    def to_json(self):  # noqa: C901

        base = {}

        if self.row_count is not None:
            base["rowCount"] = self.row_count
        if self.column_count is not None:
            base["columnCount"] = self.column_count
        if self.frozen_row_count is not None:
            base["frozenRowCount"] = self.frozen_row_count
        if self.frozen_column_count is not None:
            base["frozenColumnCount"] = self.frozen_column_count
        if self.hide_gridlines is not None:
            base["hideGridlines"] = self.hide_gridlines
        if self.row_group_control_after is not None:
            base["rowGroupControlAfter"] = self.row_group_control_after
        if self.colum_group_control_after is not None:
            base["columnGroupControlAfter"] = self.colum_group_control_after

        return base


@dataclass
class SheetProperties:
    sheet_id: Optional[int] = None
    title: Optional[str] = None
    index: Optional[int] = None
    sheet_type: Optional[SheetType] = None
    grid_properties: Optional[GridProperties] = None
    hidden: Optional[bool] = None
    tab_color: Optional[Color] = None
    # TODO: tab_color_style
    right_to_left: Optional[bool] = None

    def __post_init__(self):
        self.tab_color = Color.parse(self.tab_color)

        # sheetId must be non-negative
        if self.sheet_id:
            assert self.sheet_id >= 0

    def to_json(self):  # noqa: C901

        base = {}

        if self.sheet_id is not None:
            base["sheetId"] = self.sheet_id
        if self.title is not None:
            base["title"] = self.title
        if self.index is not None:
            base["index"] = self.index
        if self.sheet_type is not None:
            base["sheetType"] = self.sheet_type
        if self.grid_properties is not None:
            base["gridProperties"] = self.grid_properties.to_json()
        if self.hidden is not None:
            base["hidden"] = self.hidden
        if self.tab_color is not None:
            base["tabColor"] = self.tab_color.to_sheet_style()
        # TODO: if self.tab_color_style is not None:
        #     base["tabColorStyle"] = self.tab_color_style
        if self.right_to_left is not None:
            base["rightToLeft"] = self.right_to_left

        return base



from typing import List, Optional

from pygsuite import Clients
from pygsuite.common.comment import Comment
from pygsuite.enums import FileTypes, PermissionType
from pygsuite.utility.decorators import lazy_property


def default_callback(request_id, response, exception):
    if exception:
        # Handle error
        raise ValueError(exception)


class DriveObject:
    file_type: FileTypes = FileTypes.UNKNOWN

    def __init__(
        self,
        id,
        client,
    ):
        self.id = id
        self.client = client

    @property
    def comments(self) -> List[Comment]:
        return [
            Comment(item)
            for item in self.drive_client.comments().list(fileId=self.id).execute().get("items", [])
        ]

    @lazy_property
    def drive_client(self):
        return Clients.drive_client_v2

    @lazy_property
    def drive_client_3(self):
        return Clients.drive_client_v3

    @classmethod
    def create_new(cls, title: str, client=None):
        raise NotImplementedError

    @classmethod
    def get_safe(cls, title: str, drive_client=None, object_client=None):
        from pygsuite.drive import Drive

        file_client = drive_client or Clients.drive_client_v3
        files = Drive(client=file_client).find_files(cls.file_type, name=title)
        if files:
            return cls(id=files[0]["id"], client=object_client)
        else:
            return cls.create_new(title, object_client)

    def share(
        self,
        role: PermissionType,
        user: Optional[str] = None,
        group: Optional[str] = None,
        domain: Optional[str] = None,
        everyone: bool = False,
    ):
        """Share the object with a provided permission with a user, group, domain, or everyone."""
        permissions: List[dict] = []
        role = PermissionType(role)
        if user:
            permissions.append({"role": role.value, "type": "user", "emailAddress": user})
        if group:
            permissions.append({"role": role.value, "type": "group", "emailAddress": group})
        if domain:
            permissions.append({"role": role.value, "type": "domain", "domain": domain})
        if everyone:
            permissions.append({"role": role.value, "type": "everyone"})
        for permission in permissions:
            self.drive_client_3.permissions().create(
                fileId=self.id, body=permission, supportsAllDrives=True
            ).execute()



from os import environ, path


# AUTH
def get_credentials():
    dirname = path.dirname(path.dirname(__file__))
    return path.join(dirname, "credentials.json")


CREDENTIALS = environ.get("GOOGLE_APPLICATION_CREDENTIALS") or get_credentials()

# COMMON

# SHEETS
SHEETS_MAX_COLUMN_NUMBER = 18278

# DOCS

# SLIDES



from dataclasses import dataclass
from typing import Dict
from enum import Enum

from .base_element import BaseElement


class Chart(BaseElement):
    @classmethod
    def from_id(cls, id, presentation):
        return cls(element={"sheetsChart": {}, "objectId": id}, presentation=presentation)

    def __init__(self, element, presentation):
        BaseElement.__init__(self, element, presentation)
        self._details = self._element.get("sheetsChart")

    @property
    def id(self):
        return self._element["objectId"]

    def __repr__(self):
        return f"<Chart >"

    @property
    def spreadsheet_id(self):
        return self._details.get("spreadsheetId")

    @property
    def chart_id(self):
        return self._details.get("chartId")

    @property
    def content_url(self):
        return self._details.get("content_url")



from typing import Dict, Union, Optional

from googleapiclient.errors import HttpError

from pygsuite import Clients
from pygsuite.common.drive_object import DriveObject
from pygsuite.common.parsing import parse_id
from pygsuite.enums import FileTypes
from pygsuite.utility.decorators import retry
from .layout import Layout
from .slide import Slide


class Presentation(DriveObject):
    file_type = FileTypes.SLIDES

    def __init__(self, id, client=None):
        from pygsuite import Clients

        self.service = client or Clients.slides_client
        self.id = parse_id(id) if id else None
        DriveObject.__init__(self, id=id, client=client)
        self._presentation = self.service.presentations().get(presentationId=self.id).execute()
        self._change_queue = []

    @classmethod
    def create_new(cls, title: str, client=None):
        client = client or Clients.slides_client
        body = {"title": title}
        new = client.presentations().create(body=body).execute()
        return Presentation(id=new.get("presentationId"), client=client)

    def __getitem__(self, item) -> Slide:
        return self.slides[item]

    def __setitem__(self, index, value: Slide, style=None):
        self._mutation([{"slideObjectIds": [Slide.id], "insertionIndex": index}])
        self.flush()

    @retry((HttpError), tries=3, delay=5, backoff=3)
    def flush(self, reverse=False):
        if reverse:
            base = reversed(self._change_queue)
        else:
            base = self._change_queue
        final = []
        for item in base:
            if isinstance(item, list):
                for i in item:
                    final.append(i)
            else:
                final.append(item)
        if not base:
            return []
        out = (
            self.service.presentations()
            .batchUpdate(body={"requests": final}, presentationId=self.id)
            .execute()["replies"]
        )

        self._change_queue = []
        self.refresh()
        return out

    def refresh(self):
        self._presentation = self.service.presentations().get(presentationId=self.id).execute()

    @property
    def slides(self):
        return [Slide(slide, self) for slide in self._presentation.get("slides")]

    def get_slide(self, id: str) -> Slide:
        return Slide.from_id(id, self)

    @property
    def layouts(self) -> Dict[str, Layout]:
        layouts = [Layout(layout, self) for layout in self._presentation.get("layouts")]
        return {layout.display_name: layout for layout in layouts}

    def _mutation(self, reqs, flush=False):
        if not reqs:
            return None
        self._change_queue += reqs
        if flush:
            return self.flush()

    @property
    def url(self):
        return f"https://docs.google.com/presentation/d/{self.id}"

    def add_slide(
        self,
        layout: Union[str, Layout] = None,
        placeholders: Dict = None,
        index: Optional[int] = None,
        flush: bool = False,
    ) -> Optional[Slide]:
        base = {}
        placeholder_mappings = []
        if layout:
            if isinstance(layout, Layout):
                id = layout.id
            else:
                id = layout
            base["slideLayoutReference"] = {"layoutId": id}
        if not index is None:  # noqa: E714
            base["insertionIndex"] = index
        if placeholders is not None:
            if not layout:
                raise ValueError("Cannot pass placeholders without a reference layout!")
            placeholder_mappings = [
                layout.build_reference_map(key, val) for key, val in placeholders.items()
            ]
            base["placeholderIdMappings"] = [map[1] for map in placeholder_mappings]
        reqs = [{"createSlide": base}]
        base_idx = -1
        for map in placeholder_mappings:
            reqs.append(map[0])
            base_idx -= 1

        # return added slide ID
        # always flush
        out = self._mutation([reqs], flush=flush)
        if flush:
            return self.get_slide(out[base_idx]["createSlide"]["objectId"])
        return None
        # self.refresh()
        # return self.get_slide(created)


#     def replace(self, args):
#         reqs = [
#             {'replaceAllText': {
#                 'containsText': {'text': '{{NAME}}'},
#                 'replaceText': 'Hello World!'
#             }},
#             {'createImage': {
#                 'url': img_url,
#                 'elementProperties': {
#                     'pageObjectId': slide['objectId'],
#                     'size': obj['size'],
#                     'transform': obj['transform'],
#                 }
#             }},
#             {'deleteObject': {'objectId': obj['objectId']}},
#         ]
# SLIDES.presentations().batchUpdate(body={'requests': reqs},
#                                    presentationId=DECK_ID).execute()



from dataclasses import dataclass
from typing import List, Union

from jinja2 import Template, Environment, meta

from .page_element import PageElement, Table, Shape, Image, Line
from .page_elements import LineProperties
from pygsuite.slides.element_properties import ElementProperties
from pygsuite.slides.enums import ShapeType
from pygsuite.utility.guids import get_guid


@dataclass
class FancyFont:
    lines: List
    font_size: float
    bullets: bool


class Slide(object):
    @classmethod
    def from_id(cls, id, presentation):
        slides = presentation.slides
        return [slide for slide in slides if slide.id == id][0]

    def __init__(self, slide, presentation):
        # TODO
        self._slide = slide
        self._properties = slide["slideProperties"]
        self._presentation = presentation

    @property
    def elements(self):
        elements = self._slide.get("pageElements")
        if not elements:
            return []
        return [PageElement(element, self._presentation) for element in elements]

    @property
    def id(self):
        return self._slide["objectId"]

    @property
    def tables(self):
        return [el for el in self.elements if isinstance(el, Table)]

    @property
    def shapes(self):
        return [el for el in self.elements if isinstance(el, Shape)]

    @property
    def images(self):
        return [el for el in self.elements if isinstance(el, Image)]

    def duplicate(self, append=True):
        # create a slide at the end of the presentation
        reqs = [{"duplicateObject": {"objectId": self.id}}]
        output = self._presentation._mutation(reqs=reqs, flush=True)[-1]["duplicateObject"][
            "objectId"
        ]

        if append:
            reqs = [
                {
                    "updateSlidesPosition": {
                        "slideObjectIds": [output],
                        "insertionIndex": len(self._presentation.slides),
                    }
                }
            ]
            self._presentation._mutation(reqs=reqs, flush=True)
        return output

    def template_text(self, dict):
        reqs = []
        for element in self.elements:
            for object in element.children:
                if object.text and any([key in object.text for key in dict]):
                    text = object.text
                    text = text.replace("’", "'")
                    text = text.replace("‘", "'")
                    env = Environment()
                    ast = env.parse(text)
                    vars = meta.find_undeclared_variables(ast)
                    if vars:
                        bullets = False
                        font_size = 8
                        render_dict = {key: dict[key] for key in dict if key in vars}
                        for var in render_dict:
                            # turn it into paragraphs if list
                            # and add bullets
                            if isinstance(render_dict[var], FancyFont):
                                ff_item = render_dict[var]
                                bullets = ff_item.bullets
                                font_size = ff_item.font_size
                                render_dict[var] = "\n".join(ff_item.lines)
                        new_text = Template(text).render(**render_dict)
                        if new_text != object.text:
                            reqs.append(
                                object.update_text(new_text, bullets=bullets, font_size=font_size)
                            )

        self._presentation._mutation(reqs=reqs)

    def add_image(self, url: str, properties: ElementProperties, id=None):
        id = id or get_guid()
        reqs = []

        reqs.append(
            {
                "createImage": {
                    "objectId": id,
                    "elementProperties": properties.to_slides_json(self.id),
                    "url": url,
                }
            }
        )
        self._presentation._mutation(reqs=reqs)
        return Image.from_id(id=id, presentation=self._presentation)

    #
    # def add_text(self, text):
    #     raise NotImplementedError

    def add_line(
        self,
        category,
        properties: ElementProperties,
        line_properties: LineProperties = None,
        id=None,
    ):
        id = id or get_guid()
        reqs = []

        reqs.append(
            {
                "createLine": {
                    "objectId": id,
                    "elementProperties": properties.to_slides_json(self.id),
                    "category": category,
                }
            }
        )
        if line_properties:
            reqs.append(line_properties.to_api_repr(id))
        self._presentation._mutation(reqs=reqs)
        return Line.from_id(id=id, presentation=self._presentation)

    def add_word_art(self, text):
        raise NotImplementedError

    def add_table(self, table):
        raise NotImplementedError

    def add_shape(
        self, shape: Union[str, ShapeType], properties: ElementProperties = None, id=None, **kwargs
    ):
        if not properties:
            properties = ElementProperties(**kwargs)
        id = id or get_guid()
        reqs = []
        if isinstance(shape, ShapeType):
            shape = shape.name
        reqs.append(
            {
                "createShape": {
                    "objectId": id,
                    "elementProperties": properties.to_slides_json(self.id),
                    "shapeType": shape,
                }
            }
        )
        self._presentation._mutation(reqs=reqs)
        return Shape.from_id(id=id, presentation=self._presentation)



import re
from typing import List, Optional, TYPE_CHECKING

import pandas as pd

from pygsuite.common.style import Border
from pygsuite.sheets.cell import Cell
from pygsuite.sheets.helpers import alphabet_to_index, index_to_alphabet

if TYPE_CHECKING:
    from pygsuite.sheets import Spreadsheet


# INDEX_SPLITTER = re.compile('(\d+)',s)


class Worksheet(object):
    """Worksheet object for the worksheets within a Spreadsheet"""

    def __init__(self, worksheet, spreadsheet: "Spreadsheet"):
        """Method to initialize the class.

        Args:
            worksheet ():
            spreadsheet (pygsuite.sheets.Spreadsheet): the Google Spreadsheet object containing the Worksheet.
        """

        self._worksheet = worksheet
        self._spreadsheet: "Spreadsheet" = spreadsheet
        self._properties = self._worksheet["properties"]

    def __getitem__(self, cell_range):
        return self.values_from_range(cell_range)

    def flush(self):
        self._spreadsheet.flush()
        return self

    @property
    def name(self):
        return self._properties["title"]

    @property
    def id(self):
        return self._properties["sheetId"]

    @property
    def row_count(self):
        return self._properties["gridProperties"]["rowCount"]

    @property
    def column_count(self):
        return self._properties["gridProperties"]["columnCount"]

    def clear(self):
        """Method to clear all records in the sheet"""
        worksheet_range = self.range_from_indexes(1, 1, self.column_count, self.row_count)
        self._spreadsheet.clear_range(range=worksheet_range)
        return self

    def delete_sheet(self, flush: Optional[bool] = True):
        self._spreadsheet.delete_sheet(id=self.id)
        if flush:
            self._spreadsheet.flush()

    def range_from_indexes(self, startcol: int, startrow: int, endcol: int, endrow: int):

        assert startcol <= endcol
        assert startrow <= endrow

        start_index = f"{index_to_alphabet(startcol)}{startrow}"
        end_index = f"{index_to_alphabet(endcol)}{endrow}"
        range_label = f"{self.name}!{start_index}:{end_index}"

        return range_label

    def values_from_range(self, cell_range: str):

        worksheet_range = f"{self.name}!{cell_range}"
        values = self._spreadsheet.get_values_from_range(worksheet_range).to_list()
        return values

    @property
    def values(self):

        worksheet_range = self.range_from_indexes(1, 1, self.column_count, self.row_count)
        values = self._spreadsheet.get_values_from_range(worksheet_range).to_list()
        return values

    @property
    def dataframe(self):

        worksheet_range = self.range_from_indexes(1, 1, self.column_count, self.row_count)
        df = self._spreadsheet.get_values_from_range(worksheet_range).to_df()
        return df

    def format_borders(
        self,
        start_row_index: int,
        end_row_index: int,
        start_column_index: int,
        end_column_index: int,
        borders: List[Border],
    ):

        request = {
            "updateBorders": {
                "range": {
                    "sheetId": self.id,
                    "startRowIndex": start_row_index,
                    "endRowIndex": end_row_index,
                    "startColumnIndex": start_column_index,
                    "endColumnIndex": end_column_index,
                }
            }
        }

        for border in borders:
            request["updateBorders"][border.position.value] = border.to_json()

        self._spreadsheet._spreadsheets_update_queue.append(request)

    def format_cells(
        self,
        start_row_index: int,
        end_row_index: int,
        start_column_index: int,
        end_column_index: int,
        cell: Cell,
    ):

        fields, cell_json = cell.to_json()

        request = {
            "repeatCell": {
                "range": {
                    "sheetId": self.id,
                    "startRowIndex": start_row_index,
                    "endRowIndex": end_row_index,
                    "startColumnIndex": start_column_index,
                    "endColumnIndex": end_column_index,
                },
                "cell": cell_json,
                "fields": fields,
            }
        }

        self._spreadsheet._spreadsheets_update_queue.append(request)

    def insert_data(
        self,
        values: list,
        insert_range: str = None,
        anchor: str = None,
        flush: bool = False,
        # major_dimension: Dimension = Dimension.ROWS
        # commented out because you'll need to move this to avoid circular imports
    ):
        # DO some input validation here
        anchor = insert_range.split(":")[0] if insert_range else anchor if anchor else "A1"

        split = re.split(r"(\d+)", anchor)
        x_raw = split[0]
        x = alphabet_to_index(x_raw)
        y = int(split[1])

        values_x = x + len(values[0])
        values_y = len(values) + y

        top_left = anchor
        bottom_right = index_to_alphabet(values_x) + str(values_y)

        # do some validation that it fits in rage if insert_rang was provided

        range = f"{self.name}!{top_left}:{bottom_right}"
        self._spreadsheet.insert_data(insert_range=range, values=values)

        if flush:
            self._spreadsheet.flush()

    def insert_data_from_df(
        self,
        df: pd.DataFrame,
        header: Optional[bool] = True,
        insert_range: Optional[str] = None,
        anchor: Optional[str] = None,
        flush: bool = False,
    ):

        values = []
        if header:
            values.append(df.columns.values.tolist())
        values.extend(df.values.tolist())

        self.insert_data(values=values, insert_range=insert_range, anchor=anchor, flush=flush)
